# coding=utf-8
# Copyright 2020 The HuggingFace NLP Authors and the TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""ClassLabel feature."""

import os

import pyarrow as pa

from .feature import Tensor


class ClassLabel(Tensor):
	"""`FeatureConnector` for integer class labels."""

	def __init__(self, num_classes=None, names=None, names_file=None):
		"""Constructs a ClassLabel FeatureConnector.

		There are 3 ways to define a ClassLabel, which correspond to the 3
		arguments:

		 * `num_classes`: create 0 to (num_classes-1) labels
		 * `names`: a list of label strings
		 * `names_file`: a file containing the list of labels.

		Note: On python2, the strings are encoded as utf-8.

		Args:
			num_classes: `int`, number of classes. All labels must be < num_classes.
			names: `list<str>`, string names for the integer classes. The
				order in which the names are provided is kept.
			names_file: `str`, path to a file with names for the integer
				classes, one per line.
		"""
		super(ClassLabel, self).__init__(shape=(), dtype=pa.int64)

		self._num_classes = None
		self._str2int = None
		self._int2str = None

		# The label is explicitly set as undefined (no label defined)
		if not sum(bool(a) for a in (num_classes, names, names_file)):
			return

		if sum(bool(a) for a in (num_classes, names, names_file)) != 1:
			raise ValueError(
					"Only a single argument of ClassLabel() should be provided.")

		if num_classes:
			self._num_classes = num_classes
		else:
			self.names = names or _load_names_from_file(names_file)

	@property
	def num_classes(self):
		return self._num_classes

	@property
	def names(self):
		if not self._int2str:
			return [str(i) for i in range(self._num_classes)]
		return list(self._int2str)

	@names.setter
	def names(self, new_names):
		int2str = [str(name) for name in new_names]
		# Names can only be defined once
		if self._int2str is not None and self._int2str != int2str:
			raise ValueError(
					"Trying to overwrite already defined ClassLabel names. Previous: {} "
					", new: {}".format(self._int2str, int2str))

		# Set-up [new] names
		self._int2str = int2str
		self._str2int = {name: i for i, name in enumerate(self._int2str)}
		if len(self._int2str) != len(self._str2int):
			raise ValueError(
					"Some label names are duplicated. Each label name should be unique.")

		# If num_classes has been defined, ensure that num_classes and names match
		num_classes = len(self._str2int)
		if self._num_classes is None:
			self._num_classes = num_classes
		elif self._num_classes != num_classes:
			raise ValueError(
					"ClassLabel number of names do not match the defined num_classes. "
					"Got {} names VS {} num_classes".format(
							num_classes, self._num_classes)
			)

	def str2int(self, str_value):
		"""Conversion class name string => integer."""
		str_value = str(str_value)
		if self._str2int:
			return self._str2int[str_value]

		# No names provided, try to integerize
		failed_parse = False
		try:
			int_value = int(str_value)
		except ValueError:
			failed_parse = True
		if failed_parse or not 0 <= int_value < self._num_classes:
			raise ValueError("Invalid string class label %s" % str_value)
		return int_value

	def int2str(self, int_value):
		"""Conversion integer => class name string."""
		if self._int2str:
			# Maybe should support batched np array/eager tensors, to allow things
			# like
			# out_ids = model(inputs)
			# labels = cifar10.info.features['label'].int2str(out_ids)
			return self._int2str[int_value]

		# No names provided, return str(int)
		if not 0 <= int_value < self._num_classes:
			raise ValueError("Invalid integer class label %d" % int_value)
		return str(int_value)

	def encode_example(self, example_data):
		if self._num_classes is None:
			raise ValueError(
					"Trying to use ClassLabel feature with undefined number of class. "
					"Please set ClassLabel.names or num_classes."
			)

		# If a string is given, convert to associated integer
		if isinstance(example_data, str):
			example_data = self.str2int(example_data)

		# Allowing -1 to mean no label.
		if not -1 <= example_data < self._num_classes:
			raise ValueError("Class label %d greater than configured num_classes %d" %
											 (example_data, self._num_classes))
		return example_data

	def save_metadata(self, data_dir, feature_name=None):
		"""See base class for details."""
		# Save names if defined
		if self._str2int is not None:
			names_filepath = _get_names_filepath(data_dir, feature_name)
			_write_names_to_file(names_filepath, self.names)

	def load_metadata(self, data_dir, feature_name=None):
		"""See base class for details."""
		# Restore names if defined
		names_filepath = _get_names_filepath(data_dir, feature_name)
		if os.path.exists(names_filepath):
			self.names = _load_names_from_file(names_filepath)

	def _additional_repr_info(self):
		return {"num_classes": self.num_classes}


def _get_names_filepath(data_dir, feature_name):
	return os.path.join(data_dir, "{}.labels.txt".format(feature_name))


def _load_names_from_file(names_filepath):
	with open(names_filepath, "r") as f:
		return [
				name.strip()
				for name in f.read().split("\n")
				if name.strip()  # Filter empty names
		]


def _write_names_to_file(names_filepath, names):
	with open(names_filepath, "w") as f:
		f.write("\n".join(names) + "\n")



# coding=utf-8
# Copyright 2020 The HuggingFace NLP Authors and the TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Access datasets."""

import abc
import inspect
import re
import os
import json
import shutil
import logging
import importlib

from typing import Optional, Dict, Union

from . import naming
from .builder import DatasetBuilder
from .utils import py_utils
from .splits import Split
from .utils.file_utils import (HF_DATASETS_CACHE, is_remote_url, hf_bucket_url,
                         cached_path, url_to_filename)

logger = logging.getLogger(__name__)

__all__ = [
        "builder",
        "load",
]


CURRENT_FILE_DIRECTORY = os.path.dirname(os.path.abspath(__file__))
DATASETS_PATH = os.path.join(CURRENT_FILE_DIRECTORY, 'datasets')
DATASETS_MODULE = "nlp.datasets"


def load_dataset(path: str,
                 name: Optional[str] = None,
                 force_reload: bool = False,
                 resume_download: bool = False,
                 proxies: Optional[Dict] = None,
                 local_files_only: bool = False,
                 data_dir: Optional[str] = None,
                 **kwargs):
    r"""
        Download/extract/cache a dataset to add to the lib from a path or url which can be:
            - a path to a local directory containing the dataset processing python script
            - an url to a S3 directory with a dataset processing python script

        Dataset codes are cached inside the lib to allow easy import (avoid ugly sys.path tweaks)
        and using cloudpickle (among other things).

        Return: tuple of
            the unique id associated to the dataset
            the local path to the dataset
    """
    if name is None:
        name = list(filter(lambda x: x, path.split('/')))[-1] + '.py'

    combined_path = os.path.join(path, name)
    if os.path.isfile(path) or is_remote_url(path):
        dataset_file = path
    elif os.path.isfile(combined_path) or is_remote_url(combined_path):
        dataset_file = combined_path
    else:
        dataset_file = hf_bucket_url(path, postfix=name)

    # Get the file on our local file system (either cache_dir or already local path)
    local_path = cached_path(
        dataset_file,
        cache_dir=data_dir,
        force_download=force_reload,
        extract_compressed_file=True,
        force_extract=force_reload,
        local_files_only=local_files_only,
    )

    # Define a directory with a unique name in our dataset folder
    dataset_id = url_to_filename(local_path)
    dataset_folder_path = os.path.join(DATASETS_PATH, dataset_id)
    dataset_file_path = os.path.join(dataset_folder_path, name)

    # Check if the dataset directory was already there
    if os.path.exists(dataset_file_path) and not force_reload:
        logger.info("Dataset script %s already found in datasets directory at %s, returning it. Use `force_reload=True` to override.",
                    local_path, dataset_file_path)
    else:
        # Create directory for dataset in DATASETS_PATH
        logger.info("Creating unique folder for dataset %s in datasets directory at %s", local_path, dataset_folder_path)
        shutil.rmtree(dataset_folder_path, ignore_errors=True)
        os.makedirs(dataset_folder_path)

        # Copy dataset.py file there
        shutil.copyfile(local_path, dataset_file_path)

        # Record metadata associating original dataset path with local unique folder
        logger.info("Creating metadata file for dataset %s at %s", local_path, dataset_file_path + ".json")
        meta = {"original path": local_path, "library path": dataset_file_path}
        meta_path = dataset_file_path + ".json"
        with open(meta_path, "w") as meta_file:
            json.dump(meta, meta_file)

        # Add an empty __init__ file to load the module
        init_file_path = os.path.join(dataset_folder_path, '__init__.py')
        with open(init_file_path, 'w'):
            pass

    importlib.invalidate_caches()
    module_name = name.replace('.py', '')
    module_path = '.'.join([DATASETS_MODULE, dataset_id, module_name])
    dataset_module = importlib.import_module(module_path)

    builder_cls = None
    for name, obj in dataset_module.__dict__.items():
        if isinstance(obj, type) and issubclass(obj, DatasetBuilder):
            builder_cls = obj
            builder_cls.name = naming.camelcase_to_snakecase(name)
            break

    return builder_cls


def builder(path: str,
            name: Optional[str] = None,
            **builder_init_kwargs):
    """Fetches a `nlp.DatasetBuilder` by string name.

    Args:
        name: `str`, the name of the `DatasetBuilder` (the snake case
            version of the class name). This can be either `"dataset_name"` or
            `"dataset_name/config_name"` for datasets with `BuilderConfig`s.
            As a convenience, this string may contain comma-separated keyword
            arguments for the builder. For example `"foo_bar/a=True,b=3"` would use
            the `FooBar` dataset passing the keyword arguments `a=True` and `b=3`
            (for builders with configs, it would be `"foo_bar/zoo/a=True,b=3"` to
            use the `"zoo"` config and pass to the builder keyword arguments `a=True`
            and `b=3`).
        **builder_init_kwargs: `dict` of keyword arguments passed to the
            `DatasetBuilder`. These will override keyword arguments passed in `name`,
            if any.

    Returns:
        A `nlp.DatasetBuilder`.

    Raises:
        DatasetNotFoundError: if `name` is unrecognized.
    """
    if name is not None:
        name, builder_kwargs = _dataset_name_and_kwargs_from_name_str(name)
        builder_kwargs.update(builder_init_kwargs)
    else:
        builder_kwargs = builder_init_kwargs
    builder_cls = load_dataset(path, name=name, **builder_kwargs)
    builder_instance = builder_cls(**builder_kwargs)
    return builder_instance


def load(path: str,
         name: Optional[str] = None,
         split: Optional[Union[str, Split]] = None,
         data_dir: Optional[str] = None,
                 batch_size=None,
                 in_memory=None,
                 download=True,
                 as_supervised=False,
                 with_info=False,
                 builder_kwargs=None,
                 download_and_prepare_kwargs=None,
                 as_dataset_kwargs=None):
    # pylint: disable=line-too-long
    """Loads the named dataset.

    If `split=None` (the default), returns all splits for the dataset. Otherwise,
    returns the specified split.

    `load` is a convenience method that fetches the `nlp.DatasetBuilder` by
    string name, optionally calls `DatasetBuilder.download_and_prepare`
    (if `download=True`), and then calls `DatasetBuilder.as_dataset`.
    This is roughly equivalent to:

    ```
    builder = nlp.builder(name, data_dir=data_dir, **builder_kwargs)
    if download:
        builder.download_and_prepare(**download_and_prepare_kwargs)
    ds = builder.as_dataset(
            split=split, as_supervised=as_supervised, **as_dataset_kwargs)
    if with_info:
        return ds, builder.info
    return ds
    ```

    If you'd like NumPy arrays instead of `tf.data.Dataset`s or `tf.Tensor`s,
    you can pass the return value to `nlp.as_numpy`.

    Callers must pass arguments as keyword arguments.

    **Warning**: calling this function might potentially trigger the download
    of hundreds of GiB to disk. Refer to the `download` argument.

    Args:
        name: `str`, the name of the `DatasetBuilder` (the snake case
            version of the class name). This can be either `"dataset_name"` or
            `"dataset_name/config_name"` for datasets with `BuilderConfig`s.
            As a convenience, this string may contain comma-separated keyword
            arguments for the builder. For example `"foo_bar/a=True,b=3"` would use
            the `FooBar` dataset passing the keyword arguments `a=True` and `b=3`
            (for builders with configs, it would be `"foo_bar/zoo/a=True,b=3"` to
            use the `"zoo"` config and pass to the builder keyword arguments `a=True`
            and `b=3`).
        split: `nlp.Split` or `str`, which split of the data to load. If None,
            will return a `dict` with all splits (typically `nlp.Split.TRAIN` and
            `nlp.Split.TEST`).
        data_dir: `str` (optional), directory to read/write data.
            Defaults to "~/nlp".
        batch_size: `int`, if set, add a batch dimension to examples. Note that
            variable length features will be 0-padded. If
            `batch_size=-1`, will return the full dataset as `tf.Tensor`s.
        in_memory: `bool`, if `True`, loads the dataset in memory which
            increases iteration speeds. Note that if `True` and the dataset has
            unknown dimensions, the features will be padded to the maximum
            size across the dataset.
        download: `bool` (optional), whether to call
            `nlp.DatasetBuilder.download_and_prepare`
            before calling `tf.DatasetBuilder.as_dataset`. If `False`, data is
            expected to be in `data_dir`. If `True` and the data is already in
            `data_dir`, `download_and_prepare` is a no-op.
        as_supervised: `bool`, if `True`, the returned `tf.data.Dataset`
            will have a 2-tuple structure `(input, label)` according to
            `builder.info.supervised_keys`. If `False`, the default,
            the returned `tf.data.Dataset` will have a dictionary with all the
            features.
        with_info: `bool`, if True, nlp.load will return the tuple
            (tf.data.Dataset, nlp.DatasetInfo) containing the info associated
            with the builder.
        builder_kwargs: `dict` (optional), keyword arguments to be passed to the
            `nlp.DatasetBuilder` constructor. `data_dir` will be passed
            through by default.
        download_and_prepare_kwargs: `dict` (optional) keyword arguments passed to
            `nlp.DatasetBuilder.download_and_prepare` if `download=True`. Allow
            to control where to download and extract the cached data. If not set,
            cache_dir and manual_dir will automatically be deduced from data_dir.
        as_dataset_kwargs: `dict` (optional), keyword arguments passed to
            `nlp.DatasetBuilder.as_dataset`.

    Returns:
        ds: `tf.data.Dataset`, the dataset requested, or if `split` is None, a
            `dict<key: nlp.Split, value: nlp.data.Dataset>`. If `batch_size=-1`,
            these will be full datasets as `tf.Tensor`s.
        ds_info: `nlp.DatasetInfo`, if `with_info` is True, then `nlp.load`
            will return a tuple `(ds, ds_info)` containing dataset information
            (version, features, splits, num_examples,...). Note that the `ds_info`
            object documents the entire dataset, regardless of the `split` requested.
            Split-specific information is available in `ds_info.splits`.
    """
    # if name is not None:
    #     _, name_builder_kwargs = _dataset_name_and_kwargs_from_name_str(name)
    #     name_builder_kwargs.update(builder_kwargs or {})
    #     builder_kwargs = name_builder_kwargs

    if builder_kwargs is None:
        builder_kwargs = {}

    # Set data_dir
    if data_dir is None:
        data_dir = HF_DATASETS_CACHE

    dbuilder: DatasetBuilder = builder(path, name, data_dir=data_dir, **builder_kwargs)
    if download:
        download_and_prepare_kwargs = download_and_prepare_kwargs or {}
        dbuilder.download_and_prepare(**download_and_prepare_kwargs)

    if as_dataset_kwargs is None:
        as_dataset_kwargs = {}
    as_dataset_kwargs = dict(as_dataset_kwargs)
    as_dataset_kwargs.setdefault("split", split)
    as_dataset_kwargs.setdefault("as_supervised", as_supervised)
    as_dataset_kwargs.setdefault("batch_size", batch_size)

    ds = dbuilder.as_dataset(**as_dataset_kwargs)
    if with_info:
        return ds, dbuilder.info
    return ds


_VERSION_RE = r""

_NAME_REG = re.compile(
        r"^"
        r"(?P<dataset_name>\w+)"
        r"(/(?P<config>[\w\-\.]+))?"
        r"(:(?P<version>(\d+|\*)(\.(\d+|\*)){2}))?"
        r"(/(?P<kwargs>(\w+=\w+)(,\w+=[^,]+)*))?"
        r"$")


_NAME_STR_ERR = """\
Parsing builder name string {} failed.
The builder config string must be of the following format:
    dataset_name[/config_name][:version][/kwargs]

    Where:

        * dataset_name and config_name are string following python variable naming.
        * version is of the form x.y.z where {{x,y,z}} can be any digit or *.
        * kwargs is a comma list separated of arguments and values to pass to
            builder.

    Examples:
        my_dataset
        my_dataset:1.2.*
        my_dataset/config1
        my_dataset/config1:1.*.*
        my_dataset/config1/arg1=val1,arg2=val2
        my_dataset/config1:1.2.3/right=True,foo=bar,rate=1.2
"""

def _dataset_name_and_kwargs_from_name_str(name_str):
    """Extract kwargs from name str."""
    res = _NAME_REG.match(name_str)
    if not res:
        raise ValueError(_NAME_STR_ERR.format(name_str))
    name = res.group("dataset_name")
    kwargs = _kwargs_str_to_kwargs(res.group("kwargs"))
    try:
        for attr in ["config", "version"]:
            val = res.group(attr)
            if val is None:
                continue
            if attr in kwargs:
                raise ValueError("Dataset %s: cannot pass %s twice." % (name, attr))
            kwargs[attr] = val
        return name, kwargs
    except:
        logger.error(_NAME_STR_ERR.format(name_str))   # pylint: disable=logging-format-interpolation
        raise


def _kwargs_str_to_kwargs(kwargs_str):
    if not kwargs_str:
        return {}
    kwarg_strs = kwargs_str.split(",")
    kwargs = {}
    for kwarg_str in kwarg_strs:
        kwarg_name, kwarg_val = kwarg_str.split("=")
        kwargs[kwarg_name] = _cast_to_pod(kwarg_val)
    return kwargs


def _cast_to_pod(val):
    """Try cast to int, float, bool, str, in that order."""
    bools = {"True": True, "False": False}
    if val in bools:
        return bools[val]
    try:
        return int(val)
    except ValueError:
        try:
            return float(val)
        except ValueError:
            return str(val)



# coding=utf-8
# Copyright 2020 The HuggingFace NLP Authors and the TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
""" Arrow ArrowReader."""

import copy
import functools
import math
import os
import re

from typing import List, Optional
from dataclasses import dataclass

import pyarrow as pa

import numpy as np

from .utils import py_utils
from .naming import filename_for_dataset_split
from .arrow_dataset import Dataset


_BUFFER_SIZE = 8<<20  # 8 MiB per file.

_SUB_SPEC_RE = re.compile(r'''
^
 (?P<split>\w+)
 (\[
    ((?P<from>-?\d+)
     (?P<from_pct>%)?)?
    :
    ((?P<to>-?\d+)
     (?P<to_pct>%)?)?
 \])?
$
''', re.X)

_ADDITION_SEP_RE = re.compile(r'\s*\+\s*')


def _get_dataset_from_filename(filename_skip_take):
    """Returns a Dataset instance from given (filename, skip, take)."""
    filename, skip, take = (filename_skip_take['filename'],
                            filename_skip_take['skip'],
                            filename_skip_take['take'],)

    mmap = pa.memory_map(filename)
    f = pa.ipc.open_stream(mmap)
    pa_table = f.read_all()
    pa_table = pa_table.slice(skip, take)
    return pa_table


@dataclass(frozen=True)
class FileInstructions:
    """The file instructions associated with a split ReadInstruction.

    Attributes:
        num_examples: `int`, The total number of examples
        file_instructions: List[dict(filename, skip, take)], the files information.
            The filenames contains the relative path, not absolute.
            skip/take indicates which example read in the file: `ds.slice(skip, take)`
    """
    num_examples: int
    file_instructions: List[dict]


def make_file_instructions(name, split_infos, instruction):
    """Returns instructions of the split dict.

    Args:
        name: Name of the dataset.
        split_infos: `List[SplitInfo]`, Dataset splits information
        instruction: `ReadInstruction` or `str`

    Returns:
        file_intructions: FileInstructions instance
    """
    name2len = {
            info.name: info.num_examples for info in split_infos
    }
    if not isinstance(instruction, ReadInstruction):
        instruction = ReadInstruction.from_spec(instruction)
    # Create the absolute instruction (per split)
    absolute_instructions = instruction.to_absolute(name2len)

    return _make_file_instructions_from_absolutes(
            name=name,
            name2len=name2len,
            absolute_instructions=absolute_instructions,
    )


def _make_file_instructions_from_absolutes(
        name,
        name2len,
        absolute_instructions,
):
    """Returns the files instructions from the absolute instructions list."""
    # For each split, return the files instruction (skip/take)
    file_instructions = []
    num_examples = 0
    for abs_instr in absolute_instructions:
        length = name2len[abs_instr.splitname]
        if not length:
            raise ValueError(
                    'Split empty. This might means that dataset hasn\'t been generated '
                    'yet and info not restored from GCS, or that legacy dataset is used.')
        filename = filename_for_dataset_split(
                dataset_name=name,
                split=abs_instr.splitname,
                filetype_suffix='arrow')
        from_ = 0 if abs_instr.from_ is None else abs_instr.from_
        to = length if abs_instr.to is None else abs_instr.to
        num_examples += to - from_
        single_file_instructions = [{"filename": filename, "skip": from_, "take": to - from_}]
        file_instructions.extend(single_file_instructions)
    return FileInstructions(
            num_examples=num_examples,
            file_instructions=file_instructions,
    )


def _read_files(files, info):
    """Returns Dataset for given file instructions.

    Args:
        files: List[dict(filename, skip, take)], the files information.
            The filenames contain the absolute path, not relative.
            skip/take indicates which example read in the file: `ds.slice(skip, take)`
    """
    pa_batches = []
    for f_dict in files:
        pa_table: pa.Table = _get_dataset_from_filename(f_dict)
        pa_batches.extend(pa_table.to_batches())
    pa_table = pa.Table.from_batches(pa_batches)
    ds = Dataset(arrow_table=pa_table, data_files=files, info=info)
    return ds

class ArrowReader(object):
    """Build a Dataset object out of Instruction instance(s).

    This class should not typically be exposed to the user.
    """

    def __init__(self, path, info):
        """Initializes ArrowReader.

        Args:
            path (str): path where tfrecords are stored.
        """
        self._path = path
        self._info = info

    def read(
            self,
            name,
            instructions,
            split_infos,
    ):
        """Returns Dataset instance(s).

        Args:
            name (str): name of the dataset.
            instructions (ReadInstruction, List[], Dict[]): instruction(s) to read.
                Instructions can be string and will then be passed to the Instruction
                constructor as it.
            split_infos (list of SplitInfo proto): the available splits for dataset.

        Returns:
             a single Dataset instance if instruction is a single
             ReadInstruction instance. Otherwise a dict/list of Dataset
             corresponding to given instructions param shape.
        """
        def _read_instruction_to_ds(instruction):
            file_instructions = make_file_instructions(name, split_infos, instruction)
            files = file_instructions.file_instructions
            if not files:
                msg = 'Instruction "%s" corresponds to no data!' % instruction
                raise AssertionError(msg)
            return self.read_files(
                    files=tuple(files),
            )

        return py_utils.map_nested(_read_instruction_to_ds, instructions)

    def read_files(
            self,
            files,
    ):
        """Returns single Dataset instance for the set of file instructions.

        Args:
            files: List[dict(filename, skip, take)], the files information.
                The filenames contains the relative path, not absolute.
                skip/take indicates which example read in the file: `ds.skip().take()`

        Returns:
             a Dataset instance.
        """
        # Prepend path to filename
        files = copy.deepcopy(files)
        for f in files:
            f.update(filename=os.path.join(self._path, f['filename']))
        dataset = _read_files(
                files=files,
                info=self._info,
        )
        return dataset


@dataclass(frozen=True)
class _AbsoluteInstruction:
    """A machine friendly slice: defined absolute positive boundaries."""
    splitname: str  # type: Text
    from_: int  # uint (starting index).
    to: int  # uint (ending index).


@dataclass(frozen=True)
class _RelativeInstruction:
    """Represents a single parsed slicing instruction, can use % and negatives."""
    splitname: str  # type: Text
    from_: Optional[int] = None  # int (starting index) or None if no lower boundary.
    to: Optional[int] = None  # int (ending index) or None if no upper boundary.
    unit: Optional[str] = None
    rounding: Optional[str] = None

    def __post_init__(self):
        assert self.unit is None or self.unit in ['%', 'abs']
        assert self.rounding is None or self.rounding in ['closest', 'pct1_dropremainder']
        if self.unit == '%' and self.from_ is not None and abs(self.from_) > 100:
            raise AssertionError('Percent slice boundaries must be > -100 and < 100.')
        if self.unit == '%' and self.to is not None and abs(self.to) > 100:
            raise AssertionError('Percent slice boundaries must be > -100 and < 100.')


def _str_to_relative_instruction(spec):
    """Returns ReadInstruction for given string."""
    res = _SUB_SPEC_RE.match(spec)
    if not res:
        raise AssertionError('Unrecognized instruction format: %s' % spec)
    unit = '%' if res.group('from_pct') or res.group('to_pct') else 'abs'
    return ReadInstruction(
            split_name=res.group('split'),
            rounding='closest',
            from_=int(res.group('from')) if res.group('from') else None,
            to=int(res.group('to')) if res.group('to') else None,
            unit=unit,
            )


def _pct_to_abs_pct1(boundary, num_examples):
    # Using math.trunc here, since -99.5% should give -99%, not -100%.
    if num_examples < 100:
        msg = ('Using "pct1_dropremainder" rounding on a split with less than 100 '
                     'elements is forbidden: it always results in an empty dataset.')
        raise AssertionError(msg)
    return boundary * math.trunc(num_examples / 100.)


def _pct_to_abs_closest(boundary, num_examples):
    return int(round(boundary * num_examples / 100.))


def _rel_to_abs_instr(rel_instr, name2len):
    """Returns _AbsoluteInstruction instance for given RelativeInstruction.

    Args:
        rel_instr: RelativeInstruction instance.
        name2len: dict {split_name: num_examples}.
    """
    pct_to_abs = (_pct_to_abs_closest if rel_instr.rounding == 'closest'
                                else _pct_to_abs_pct1)
    split = rel_instr.splitname
    if split not in name2len:
        raise ValueError('Unknown split "{}". Should be one of {}.'.format(
                split, list(name2len)))
    num_examples = name2len[split]
    from_ = rel_instr.from_
    to = rel_instr.to
    if rel_instr.unit == '%':
        from_ = 0 if from_ is None else pct_to_abs(from_, num_examples)
        to = num_examples if to is None else pct_to_abs(to, num_examples)
    else:
        from_ = 0 if from_ is None else from_
        to = num_examples if to is None else to
    if abs(from_) > num_examples or abs(to) > num_examples:
        msg = 'Requested slice [%s:%s] incompatible with %s examples.' % (
                from_ or '', to or '', num_examples)
        raise AssertionError(msg)
    if from_ < 0:
        from_ = num_examples + from_
    elif from_ == 0:
        from_ = None
    if to < 0:
        to = num_examples + to
    elif to == num_examples:
        to = None
    return _AbsoluteInstruction(split, from_, to)


class ReadInstruction(object):
    """Reading instruction for a dataset.

    Examples of usage:

    ```
    # The following lines are equivalent:
    ds = nlp.load('mnist', split='test[:33%]')
    ds = nlp.load('mnist', split=ReadInstruction.from_spec('test[:33%]'))
    ds = nlp.load('mnist', split=ReadInstruction('test', to=33, unit='%'))
    ds = nlp.load('mnist', split=ReadInstruction(
            'test', from_=0, to=33, unit='%'))

    # The following lines are equivalent:
    ds = nlp.load('mnist', split='test[:33%]+train[1:-1]')
    ds = nlp.load('mnist', split=ReadInstruction.from_spec(
            'test[:33%]+train[1:-1]'))
    ds = nlp.load('mnist', split=(
            ReadInstruction.('test', to=33, unit='%') +
            ReadInstruction.('train', from_=1, to=-1, unit='abs')))

    # 10-fold validation:
    tests = nlp.load(
            'mnist',
            [ReadInstruction('train', from_=k, to=k+10, unit='%')
             for k in range(0, 100, 10)])
    trains = nlp.load(
            'mnist',
            [RI('train', to=k, unit='%') + RI('train', from_=k+10, unit='%')
             for k in range(0, 100, 10)])
    ```

    """

    def _init(self, relative_instructions):
        # Private initializer.
        self._relative_instructions = relative_instructions

    @classmethod
    def _read_instruction_from_relative_instructions(cls, relative_instructions):
        """Returns ReadInstruction obj initialized with relative_instructions."""
        # Use __new__ to bypass __init__ used by public API and not conveniant here.
        result = cls.__new__(cls)
        result._init(relative_instructions)  # pylint: disable=protected-access
        return result

    def __init__(self, split_name, rounding='closest', from_=None, to=None,
                             unit=None):
        """Initialize ReadInstruction.

        Args:
            split_name (str): name of the split to read. Eg: 'train'.
            rounding (str): The rounding behaviour to use when percent slicing is
                used. Ignored when slicing with absolute indices.
                Possible values:
                 - 'closest' (default): The specified percentages are rounded to the
                     closest value. Use this if you want specified percents to be as
                     much exact as possible.
                 - 'pct1_dropremainder': the specified percentages are treated as
                     multiple of 1%. Use this option if you want consistency. Eg:
                         len(5%) == 5 * len(1%).
                     Using this option, one might not be able to use the full set of
                     examples, if the number of those is not a multiple of 100.
            from_ (int):
            to (int): alternative way of specifying slicing boundaries. If any of
                {from_, to, unit} argument is used, slicing cannot be specified as
                string.
            unit (str): optional, one of:
                '%': to set the slicing unit as percents of the split size.
                'abs': to set the slicing unit as absolute numbers.
        """
        # This constructor is not always called. See factory method
        # `_read_instruction_from_relative_instructions`. Common init instructions
        # MUST be placed in the _init method.
        self._init(
                [_RelativeInstruction(split_name, from_, to, unit, rounding)])

    @classmethod
    def from_spec(cls, spec):
        """Creates a ReadInstruction instance out of a string spec.

        Args:
            spec (str): split(s) + optional slice(s) to read. A slice can be
                        specified, using absolute numbers (int) or percentages (int). E.g.
                            `test`: test split.
                            `test + validation`: test split + validation split.
                            `test[10:]`: test split, minus its first 10 records.
                            `test[:10%]`: first 10% records of test split.
                            `test[:-5%]+train[40%:60%]`: first 95% of test + middle 20% of
                                                                                     train.

        Returns:
            ReadInstruction instance.
        """
        spec = str(spec)  # Need to convert to str in case of NamedSplit instance.
        subs = _ADDITION_SEP_RE.split(spec)
        if not subs:
            raise AssertionError('No instructions could be built out of %s' % spec)
        instruction = _str_to_relative_instruction(subs[0])
        return sum([_str_to_relative_instruction(sub) for sub in subs[1:]],
                             instruction)

    def __add__(self, other):
        """Returns a new ReadInstruction obj, result of appending other to self."""
        if not isinstance(other, ReadInstruction):
            msg = 'ReadInstruction can only be added to another ReadInstruction obj.'
            raise AssertionError(msg)
        other_ris = other._relative_instructions  # pylint: disable=protected-access
        if self._relative_instructions[0].rounding != other_ris[0].rounding:
            raise AssertionError('It is forbidden to sum ReadInstruction instances '
                                                     'with different rounding values.')
        return self._read_instruction_from_relative_instructions(
                self._relative_instructions + other_ris)

    def __str__(self):
        return 'ReadInstruction(%s)' % self._relative_instructions

    def to_absolute(self, name2len):
        """Translate instruction into a list of absolute instructions.

        Those absolute instructions are then to be added together.

        Args:
            name2len: dict associating split names to number of examples.

        Returns:
            list of _AbsoluteInstruction instances (corresponds to the + in spec).
        """
        return [_rel_to_abs_instr(rel_instr, name2len)
                        for rel_instr in self._relative_instructions]



# coding=utf-8
# Copyright 2020 The HuggingFace NLP Authors and the TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Some python utils function and classes.

"""

import contextlib
import hashlib
import io
import itertools
import os
import sys
import uuid


from .file_utils import INCOMPLETE_SUFFIX

from shutil import disk_usage

import functools


# NOTE: When used on an instance method, the cache is shared across all
# instances and IS NOT per-instance.
# See
# https://stackoverflow.com/questions/14946264/python-lru-cache-decorator-per-instance
# For @property methods, use @memoized_property below.
memoize = functools.lru_cache


def size_str(size_in_bytes):
    """Returns a human readable size string.

    If size_in_bytes is None, then returns "Unknown size".

    For example `size_str(1.5 * nlp.units.GiB) == "1.50 GiB"`.

    Args:
        size_in_bytes: `int` or `None`, the size, in bytes, that we want to
            format as a human-readable size string.
    """
    if not size_in_bytes:
        return "Unknown size"

    _NAME_LIST = [("PiB", 2**50), ("TiB", 2**40), ("GiB",  2**30), ("MiB", 2**20),
                                ("KiB", 2**10)]

    size_in_bytes = float(size_in_bytes)
    for (name, size_bytes) in _NAME_LIST:
        value = size_in_bytes / size_bytes
        if value >= 1.0:
            return "{:.2f} {}".format(value, name)
    return "{} {}".format(int(size_in_bytes), "bytes")


def is_notebook():
    """Returns True if running in a notebook (Colab, Jupyter) environement."""
    # Inspired from the tfdm autonotebook code
    try:
        from IPython import get_ipython  # pylint: disable=import-outside-toplevel,g-import-not-at-top
        if "IPKernelApp" not in get_ipython().config:
            return False  # Run in a IPython terminal
    except:  # pylint: disable=bare-except
        return False
    else:
        return True


@contextlib.contextmanager
def temporary_assignment(obj, attr, value):
    """Temporarily assign obj.attr to value."""
    original = getattr(obj, attr, None)
    setattr(obj, attr, value)
    yield
    setattr(obj, attr, original)


def zip_dict(*dicts):
    """Iterate over items of dictionaries grouped by their keys."""
    for key in set(itertools.chain(*dicts)):  # set merge all keys
        # Will raise KeyError if the dict don't have the same keys
        yield key, tuple(d[key] for d in dicts)


class NonMutableDict(dict):
    """Dict where keys can only be added but not modified.

    Will raise an error if the user try to overwrite one key. The error message
    can be customized during construction. It will be formatted using {key} for
    the overwritten key.
    """

    def __init__(self, *args, **kwargs):
        self._error_msg = kwargs.pop(
                "error_msg",
                "Try to overwrite existing key: {key}",
        )
        if kwargs:
            raise ValueError("NonMutableDict cannot be initialized with kwargs.")
        super(NonMutableDict, self).__init__(*args, **kwargs)

    def __setitem__(self, key, value):
        if key in self:
            raise ValueError(self._error_msg.format(key=key))
        return super(NonMutableDict, self). __setitem__(key, value)

    def update(self, other):
        if any(k in self for k in other):
            raise ValueError(self._error_msg.format(key=set(self) & set(other)))
        return super(NonMutableDict, self).update(other)


class classproperty(property):  # pylint: disable=invalid-name
    """Descriptor to be used as decorator for @classmethods."""

    def __get__(self, obj, objtype=None):
        return self.fget.__get__(None, objtype)()


class memoized_property(property):  # pylint: disable=invalid-name
    """Descriptor that mimics @property but caches output in member variable."""

    def __get__(self, obj, objtype=None):
        # See https://docs.python.org/3/howto/descriptor.html#properties
        if obj is None:
            return self
        if self.fget is None:
            raise AttributeError("unreadable attribute")
        attr = "__cached_" + self.fget.__name__
        cached = getattr(obj, attr, None)
        if cached is None:
            cached = self.fget(obj)
            setattr(obj, attr, cached)
        return cached


def map_nested(function, data_struct, dict_only=False, map_tuple=False):
    """Apply a function recursively to each element of a nested data struct."""

    # Could add support for more exotic data_struct, like OrderedDict
    if isinstance(data_struct, dict):
        return {
                k: map_nested(function, v, dict_only, map_tuple)
                for k, v in data_struct.items()
        }
    elif not dict_only:
        types = [list]
        if map_tuple:
            types.append(tuple)
        if isinstance(data_struct, tuple(types)):
            mapped = [map_nested(function, v, dict_only, map_tuple)
                                for v in data_struct]
            if isinstance(data_struct, list):
                return mapped
            else:
                return tuple(mapped)
    # Singleton
    return function(data_struct)


def zip_nested(arg0, *args, **kwargs):
    """Zip data struct together and return a data struct with the same shape."""
    # Python 2 do not support kwargs only arguments
    dict_only = kwargs.pop("dict_only", False)
    assert not kwargs

    # Could add support for more exotic data_struct, like OrderedDict
    if isinstance(arg0, dict):
        return {
                k: zip_nested(*a, dict_only=dict_only) for k, a in zip_dict(arg0, *args)
        }
    elif not dict_only:
        if isinstance(arg0, list):
            return [zip_nested(*a, dict_only=dict_only) for a in zip(arg0, *args)]
    # Singleton
    return (arg0,) + args


def flatten_nest_dict(d):
    """Return the dict with all nested keys flattened joined with '/'."""
    # Use NonMutableDict to ensure there is no collision between features keys
    flat_dict = NonMutableDict()
    for k, v in d.items():
        if isinstance(v, dict):
            flat_dict.update({
                    "{}/{}".format(k, k2): v2 for k2, v2 in flatten_nest_dict(v).items()
            })
        else:
            flat_dict[k] = v
    return flat_dict


def pack_as_nest_dict(flat_d, nest_d):
    """Pack a 1-lvl dict into a nested dict with same structure as `nest_d`."""
    nest_out_d = {}
    for k, v in nest_d.items():
        if isinstance(v, dict):
            v_flat = flatten_nest_dict(v)
            sub_d = {
                    k2: flat_d.pop("{}/{}".format(k, k2)) for k2, _ in v_flat.items()
            }
            # Recursivelly pack the dictionary
            nest_out_d[k] = pack_as_nest_dict(sub_d, v)
        else:
            nest_out_d[k] = flat_d.pop(k)
    if flat_d:  # At the end, flat_d should be empty
        raise ValueError(
                "Flat dict strucure do not match the nested dict. Extra keys: "
                "{}".format(list(flat_d.keys())))
    return nest_out_d


def nlp_dir():
    """Path to nlp directory."""
    return os.path.dirname(os.path.dirname(os.path.dirname(__file__)))


@contextlib.contextmanager
def atomic_write(path, mode):
    """Writes to path atomically, by writing to temp file and renaming it."""
    tmp_path = "%s%s_%s" % (path, INCOMPLETE_SUFFIX, uuid.uuid4().hex)
    with open(tmp_path, mode) as file_:
        yield file_
    os.rename(tmp_path, path)


class abstractclassmethod(classmethod):  # pylint: disable=invalid-name
    """Decorate a method to mark it as an abstract @classmethod."""

    __isabstractmethod__ = True

    def __init__(self, fn):
        fn.__isabstractmethod__ = True
        super(abstractclassmethod, self).__init__(fn)


def get_nlp_path(relative_path):
    """Returns absolute path to file given path relative to nlp root."""
    path = os.path.join(nlp_dir(), relative_path)
    return path


def read_checksum_digest(path, checksum_cls=hashlib.sha256):
    """Given a hash constructor, returns checksum digest and size of file."""
    checksum = checksum_cls()
    size = 0
    with open(path, "rb") as f:
        while True:
            block = f.read(io.DEFAULT_BUFFER_SIZE)
            size += len(block)
            if not block:
                break
            checksum.update(block)
    return checksum.hexdigest(), size


def reraise(prefix=None, suffix=None):
    """Reraise an exception with an additional message."""
    exc_type, exc_value, exc_traceback = sys.exc_info()
    prefix = prefix or ""
    suffix = "\n" + suffix if suffix else ""
    msg = prefix + str(exc_value) + suffix
    raise(exc_type, exc_type(msg), exc_traceback)


@contextlib.contextmanager
def try_reraise(*args, **kwargs):
    """Reraise an exception with an additional message."""
    try:
        yield
    except Exception:   # pylint: disable=broad-except
        reraise(*args, **kwargs)


def rgetattr(obj, attr, *args):
    """Get attr that handles dots in attr name."""
    def _getattr(obj, attr):
        return getattr(obj, attr, *args)
    return functools.reduce(_getattr, [obj] + attr.split("."))


def has_sufficient_disk_space(needed_bytes, directory="."):
    try:
        free_bytes = disk_usage(os.path.abspath(directory)).free
    except OSError:
        return True
    return needed_bytes < free_bytes



# coding=utf-8
# Copyright 2020 The HuggingFace NLP Authors and the TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Translation feature that supports multiple languages."""

from .features_dict import FeaturesDict, Sequence
from .text_feature import Text
try:
    # This fallback applies for all versions of Python before 3.3
    import collections.abc as collections_abc  # pylint:disable=g-import-not-at-top
except ImportError:
    import collections as collections_abc  # pylint:disable=g-import-not-at-top


class Translation(FeaturesDict):
    """`FeatureConnector` for translations with fixed languages per example.

    Input: The Translate feature accepts a dictionary for each example mapping
        string language codes to string translations.

    Output: A dictionary mapping string language codes to translations as `Text`
        features.

    Example:
    At construction time:

    ```
    nlp.features.Translation(languages=['en', 'fr', 'de'])
    ```

    During data generation:

    ```
    yield {
            'en': 'the cat',
            'fr': 'le chat',
            'de': 'die katze'
    }
    ```

    Tensor returned by `.as_dataset()`:

    ```
    {
            'en': tf.Tensor(shape=(), dtype=tf.string, numpy='the cat'),
            'fr': tf.Tensor(shape=(), dtype=tf.string, numpy='le chat'),
            'de': tf.Tensor(shape=(), dtype=tf.string, numpy='die katze'),
    }
    ```
    """

    def __init__(self, languages, encoder=None, encoder_config=None):
        """Constructs a Translation FeatureConnector.

        Args:
            languages: `list<string>` Full list of languages codes.
            encoder: `nlp.features.text.TextEncoder` or
                list<nlp.features.text.TextEncoder> (optional), an encoder that can
                convert text to integer. One can be shared one per language provided. If
                None, the text will be utf-8 byte-encoded.
            encoder_config: `nlp.features.text.TextEncoderConfig` or
                `list<nlp.features.text.TextEncoderConfig>` (optional), needed
                if restoring from a file with `load_metadata`. One config can be shared
                or one per language can be provided.
        """
        # If encoder and encoder_config aren't lists, use the same values for all
        # languages.
        if not isinstance(encoder, collections_abc.Iterable):
            encoder = [encoder] * len(languages)
        if not isinstance(encoder_config, collections_abc.Iterable):
            encoder_config = [encoder_config] * len(languages)

        super(Translation, self).__init__(
                {lang: Text(enc, enc_conf) for lang, enc, enc_conf in zip(
                        languages, encoder, encoder_config)})

    @property
    def languages(self):
        """List of languages."""
        return sorted(self.keys())


class TranslationVariableLanguages(Sequence):
    """`FeatureConnector` for translations with variable languages per example.

    Input: The TranslationVariableLanguages feature accepts a dictionary for each
        example mapping string language codes to one or more string translations.
        The languages present may vary from example to example.

    Output:
        language: variable-length 1D tf.Tensor of tf.string language codes, sorted
            in ascending order.
        translation: variable-length 1D tf.Tensor of tf.string plain text
            translations, sorted to align with language codes.

    Example (fixed language list):
    At construction time:

    ```
    nlp.features.Translation(languages=['en', 'fr', 'de'])
    ```

    During data generation:

    ```
    yield {
            'en': 'the cat',
            'fr': ['le chat', 'la chatte,']
            'de': 'die katze'
    }
    ```

    Tensor returned by `.as_dataset()`:

    ```
    {
            'language': tf.Tensor(
                    shape=(4,), dtype=tf.string, numpy=array(['en', 'de', 'fr', 'fr']),
            'translation': tf.Tensor(
                    shape=(4,), dtype=tf.string,
                    numpy=array(['the cat', 'die katze', 'la chatte', 'le chat'])),
    }
    ```
    """

    def __init__(self, languages=None):
        """Constructs a Translation FeatureConnector.

        Args:
            languages: `list<string>` (optional), full list of language codes if known
                in advance.
        """
        # TODO(adarob): Add optional text encoders once `Sequence` adds support
        # for FixedVarLenFeatures.

        self._languages = set(languages) if languages else None
        super(TranslationVariableLanguages, self).__init__({
                "language": Text(),
                "translation": Text(),
        })

    @property
    def num_languages(self):
        """Number of languages or None, if not specified in advance."""
        return len(self._languages) if self._languages else None

    @property
    def languages(self):
        """List of languages or None, if not specified in advance."""
        return sorted(list(self._languages)) if self._languages else None

    def encode_example(self, translation_dict):
        if self.languages and set(translation_dict) - self._languages:
            raise ValueError(
                    "Some languages in example ({0}) are not in valid set ({1}).".format(
                            ", ".join(sorted(set(translation_dict) - self._languages)),
                            ", ".join(self.languages)))

        # Convert dictionary into tuples, splitting out cases where there are
        # multiple translations for a single language.
        translation_tuples = []
        for lang, text in translation_dict.items():
            if isinstance(text, str):
                translation_tuples.append((lang, text))
            else:
                translation_tuples.extend([(lang, el) for el in text])

        # Ensure translations are in ascending order by language code.
        languages, translations = zip(*sorted(translation_tuples))

        return super(TranslationVariableLanguages, self).encode_example(
                {"language": languages,
                 "translation": translations})




# coding=utf-8
# Copyright 2020 The HuggingFace NLP Authors and the TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""DatasetBuilder base class."""

import abc
import functools
import inspect
import itertools
import contextlib
import shutil
import os
import sys

import logging

from . import splits as splits_lib
from . import utils
from .download import GenerateMode, DownloadConfig, DownloadManager
from .lazy_imports_lib import lazy_imports
from .naming import filename_prefix_for_split
from .arrow_reader import ArrowReader
from .arrow_writer import ArrowWriter, BeamWriter
from .utils.file_utils import HF_DATASETS_CACHE

logger = logging.getLogger(__name__)

FORCE_REDOWNLOAD = GenerateMode.FORCE_REDOWNLOAD
REUSE_CACHE_IF_EXISTS = GenerateMode.REUSE_CACHE_IF_EXISTS
REUSE_DATASET_IF_EXISTS = GenerateMode.REUSE_DATASET_IF_EXISTS


class BuilderConfig(object):
    """Base class for `DatasetBuilder` data configuration.

    DatasetBuilder subclasses with data configuration options should subclass
    `BuilderConfig` and add their own properties.
    """

    def __init__(self, name, version=None, supported_versions=None,
                             description=None):
        self._name = name
        self._version = version
        self._supported_versions = supported_versions or []
        self._description = description

    @property
    def name(self):
        return self._name

    @property
    def version(self):
        return self._version

    @property
    def supported_versions(self):
        return self._supported_versions

    @property
    def description(self):
        return self._description

    def __repr__(self):
        return "<{cls_name} name={name}, version={version}>".format(
                cls_name=type(self).__name__,
                name=self.name,
                version=self.version or "None")

class DatasetBuilder():
    """Abstract base class for all datasets.

    `DatasetBuilder` has 3 key methods:

        * `nlp.DatasetBuilder.info`: documents the dataset, including feature
            names, types, and shapes, version, splits, citation, etc.
        * `nlp.DatasetBuilder.download_and_prepare`: downloads the source data
            and writes it to disk.
        * `nlp.DatasetBuilder.as_dataset`: generate an `Dataset`.

    **Configuration**: Some `DatasetBuilder`s expose multiple variants of the
    dataset by defining a `nlp.BuilderConfig` subclass and accepting a
    config object (or name) on construction. Configurable datasets expose a
    pre-defined set of configurations in `nlp.DatasetBuilder.builder_configs`.

    Typical `DatasetBuilder` usage:

    ```python
    mnist_builder = nlp.builder("mnist")
    mnist_info = mnist_builder.info
    mnist_builder.download_and_prepare()
    datasets = mnist_builder.as_dataset()

    train_dataset, test_dataset = datasets["train"], datasets["test"]
    assert isinstance(train_dataset, Dataset`)

    # And then the rest of your input pipeline
    train_dataset = train_dataset.repeat().shuffle(1024).batch(128)
    train_dataset = train_dataset.prefetch(2)
    features = tf.compat.v1.data.make_one_shot_iterator(train_dataset).get_next()
    image, label = features['image'], features['label']
    ```
    """

    # Name of the dataset, filled by metaclass based on class name.
    name = None

    # Semantic version of the dataset (ex: nlp.Version('1.2.0'))
    VERSION = None

    # List dataset versions which can be loaded using current code.
    # Data can only be prepared with canonical VERSION or above.
    SUPPORTED_VERSIONS = []

    # Named configurations that modify the data generated by download_and_prepare.
    BUILDER_CONFIGS = []

    # Must be set for datasets that use 'manual_dir' functionality - the ones
    # that require users to do additional steps to download the data
    # (this is usually due to some external regulations / rules).
    #
    # This field should contain a string with user instructions, including
    # the list of files that should be present. It will be
    # displayed in the dataset documentation.
    MANUAL_DOWNLOAD_INSTRUCTIONS = None


    def __init__(self, data_dir=None, config=None, version=None):
        """Constructs a DatasetBuilder.

        Callers must pass arguments as keyword arguments.

        Args:
            data_dir: `str`, directory to read/write data. Defaults to
                "~/nlp".
            config: `nlp.BuilderConfig` or `str` name, optional configuration
                for the dataset that affects the data generated on disk. Different
                `builder_config`s will have their own subdirectories and versions.
            version: `str`. Optional version at which to load the dataset. An error is
                raised if specified version cannot be satisfied. Eg: '1.2.3', '1.2.*'.
                The special value "experimental_latest" will use the highest version,
                even if not default. This is not recommended unless you know what you
                are doing, as the version could be broken.
        """
        # For pickling:
        self._original_state = dict(data_dir=data_dir, config=config, version=version)
        # To do the work:
        self._builder_config = self._create_builder_config(config)
        # Extract code version (VERSION or config)
        if not self._builder_config and not self.VERSION:
            raise AssertionError(
                    "DatasetBuilder {} does not have a defined version. Please add a "
                    "`VERSION = nlp.Version('x.y.z')` to the class.".format(
                            self.name))
        self._version = self._pick_version(version)
        self._data_dir_root = os.path.expanduser(data_dir or HF_DATASETS_CACHE)
        self._data_dir = self._build_data_dir()
        if os.path.exists(self._data_dir):
            logger.info("Overwrite dataset info from restored data version.")
            self.info.read_from_directory(self._data_dir)

    def __getstate__(self):
        return self._original_state

    def __setstate__(self, state):
        self.__init__(**state)

    @property
    def builder_config(self):
        """`nlp.BuilderConfig` for this builder."""
        return self._builder_config

    def _create_builder_config(self, builder_config):
        """Create and validate BuilderConfig object."""
        if builder_config is None and self.BUILDER_CONFIGS:
            builder_config = self.BUILDER_CONFIGS[0]
            logger.info("No config specified, defaulting to first: %s/%s", self.name,
                                     builder_config.name)
        if not builder_config:
            return None
        if isinstance(builder_config, str):
            name = builder_config
            builder_config = self.builder_configs.get(name)
            if builder_config is None:
                raise ValueError("BuilderConfig %s not found. Available: %s" %
                                                 (name, list(self.builder_configs.keys())))
        name = builder_config.name
        if not name:
            raise ValueError("BuilderConfig must have a name, got %s" % name)
        is_custom = name not in self.builder_configs
        if is_custom:
            logger.warning("Using custom data configuration %s", name)
        else:
            if builder_config is not self.builder_configs[name]:
                raise ValueError(
                        "Cannot name a custom BuilderConfig the same as an available "
                        "BuilderConfig. Change the name. Available BuilderConfigs: %s" %
                        (list(self.builder_configs.keys())))
            if not builder_config.version:
                raise ValueError("BuilderConfig %s must have a version" % name)
            if not builder_config.description:
                raise ValueError("BuilderConfig %s must have a description" % name)
        return builder_config

    @utils.classproperty
    @classmethod
    @utils.memoize()
    def builder_configs(cls):
        """Pre-defined list of configurations for this builder class."""
        config_dict = {config.name: config for config in cls.BUILDER_CONFIGS}
        if len(config_dict) != len(cls.BUILDER_CONFIGS):
            names = [config.name for config in cls.BUILDER_CONFIGS]
            raise ValueError(
                    "Names in BUILDER_CONFIGS must not be duplicated. Got %s" % names)
        return config_dict

    @utils.memoized_property
    def canonical_version(self):
        if self._builder_config:
            return self._builder_config.version
        return self.VERSION

    @utils.memoized_property
    def supported_versions(self):
        if self._builder_config:
            return self._builder_config.supported_versions
        return self.SUPPORTED_VERSIONS

    @utils.memoized_property
    def versions(self):
        """Versions (canonical + availables), in preference order."""
        return [
                utils.Version(v) if isinstance(v, str) else v
                for v in [self.canonical_version] + self.supported_versions
        ]

    def _pick_version(self, requested_version):
        """Returns utils.Version instance, or raise AssertionError."""
        if requested_version == "experimental_latest":
            return max(self.versions)
        for version in self.versions:
            if requested_version is None or version.match(requested_version):
                return version
        available_versions = [str(v) for v in self.versions]
        msg = "Dataset {} cannot be loaded at version {}, only: {}.".format(
                self.name, requested_version, ", ".join(available_versions))
        raise AssertionError(msg)

    @property
    def version(self):
        return self._version

    @property
    def data_dir(self):
        return self._data_dir

    def _relative_data_dir(self, with_version=True):
        """Relative path of this dataset in data_dir."""
        builder_data_dir = self.name
        builder_config = self._builder_config
        if builder_config:
            builder_data_dir = os.path.join(builder_data_dir, builder_config.name)
        if not with_version:
            return builder_data_dir

        version = self._version
        version_data_dir = os.path.join(builder_data_dir, str(version))
        return version_data_dir

    def _build_data_dir(self):
        """Return the data directory for the current version."""
        builder_data_dir = os.path.join(
                self._data_dir_root, self._relative_data_dir(with_version=False))
        version_data_dir = os.path.join(
                self._data_dir_root, self._relative_data_dir(with_version=True))

        def _other_versions_on_disk():
            """Returns previous versions on disk."""
            if not os.path.exists(builder_data_dir):
                return []

            version_dirnames = []
            for dir_name in os.listdir(builder_data_dir):
                try:
                    version_dirnames.append((utils.Version(dir_name), dir_name))
                except ValueError:  # Invalid version (ex: incomplete data dir)
                    pass
            version_dirnames.sort(reverse=True)
            return version_dirnames

        # Check and warn if other versions exist on disk
        version_dirs = _other_versions_on_disk()
        if version_dirs:
            other_version = version_dirs[0][0]
            if other_version != self._version:
                warn_msg = (
                        "Found a different version {other_version} of dataset {name} in "
                        "data_dir {data_dir}. Using currently defined version "
                        "{cur_version}.".format(
                                other_version=str(other_version),
                                name=self.name,
                                data_dir=self._data_dir_root,
                                cur_version=str(self._version)))
                logger.warning(warn_msg)

        return version_data_dir

    def download_and_prepare(self, download_dir=None, download_config=None):
        """Downloads and prepares dataset for reading.

        Args:
            download_dir: `str`, directory where downloaded files are stored.
                Defaults to "~/nlp/downloads".
            download_config: `nlp.download.DownloadConfig`, further configuration for
                downloading and preparing dataset.

        Raises:
            IOError: if there is not enough disk space available.
        """

        download_config = download_config or DownloadConfig()
        data_exists = os.path.exists(self._data_dir)
        if data_exists and download_config.download_mode == REUSE_DATASET_IF_EXISTS:
            logger.info("Reusing dataset %s (%s)", self.name, self._data_dir)
            return

        # Currently it's not possible to overwrite the data because it would
        # conflict with versioning: If the last version has already been generated,
        # it will always be reloaded and data_dir will be set at construction.
        if data_exists:
            raise ValueError(
                    "Trying to overwrite an existing dataset {} at {}. A dataset with "
                    "the same version {} already exists. If the dataset has changed, "
                    "please update the version number.".format(self.name, self._data_dir, self.version))

        logger.info("Generating dataset %s (%s)", self.name, self._data_dir)
        if not utils.has_sufficient_disk_space(
                self.info.dataset_size + self.info.download_size,
                directory=self._data_dir_root):
            raise IOError(
                    "Not enough disk space. Needed: {} (download: {}, generated: {})"
                    .format(
                            utils.size_str(self.info.dataset_size + self.info.download_size),
                            utils.size_str(self.info.download_size),
                            utils.size_str(self.info.dataset_size),
                    ))

        # Print is intentional: we want this to always go to stdout so user has
        # information needed to cancel download/preparation if needed.
        # This comes right before the progress bar.
        print(
                "Downloading and preparing dataset {} (download: {}, generated: {}, "
                "total: {}) to {}...".format(
                        self.info.full_name,
                        utils.size_str(self.info.download_size),
                        utils.size_str(self.info.dataset_size),
                        utils.size_str(self.info.download_size + self.info.dataset_size),
                        self._data_dir,
                ))

        dl_manager = self._make_download_manager(
                download_dir=download_dir,
                download_config=download_config)

        @contextlib.contextmanager
        def incomplete_dir(dirname):
            """Create temporary dir for dirname and rename on exit."""
            tmp_dir = dirname + ".incomplete"
            os.makedirs(tmp_dir)
            try:
                yield tmp_dir
                os.rename(tmp_dir, dirname)
            finally:
                if os.path.exists(tmp_dir):
                    shutil.rmtree(tmp_dir)

        # Create a tmp dir and rename to self._data_dir on successful exit.
        with incomplete_dir(self._data_dir) as tmp_data_dir:
            # Temporarily assign _data_dir to tmp_data_dir to avoid having to forward
            # it to every sub function.
            with utils.temporary_assignment(self, "_data_dir", tmp_data_dir):
                self._download_and_prepare(
                        dl_manager=dl_manager,
                        download_config=download_config)

                # NOTE: If modifying the lines below to put additional information in
                # DatasetInfo, you'll likely also want to update
                # DatasetInfo.read_from_directory to possibly restore these attributes
                # when reading from package data.

                splits = list(self.info.splits.values())
                # statistics_already_computed = bool(
                #         splits and splits[0].num_examples)
                # # Update DatasetInfo metadata by computing statistics from the data.
                # if (download_config.compute_stats == download.ComputeStatsMode.SKIP or
                #         download_config.compute_stats == download.ComputeStatsMode.AUTO
                #         and statistics_already_computed
                #         ):
                #     logger.info(
                #             "Skipping computing stats for mode %s.",
                #             download_config.compute_stats)
                # else:  # Mode is forced or stats do not exists yet
                #     logger.info("No statistics computed for now.")
                    # self.info.compute_dynamic_properties()
                self.info.download_size = dl_manager.downloaded_size
                # Write DatasetInfo to disk, even if we haven't computed statistics.
                self.info.write_to_directory(self._data_dir)

        msg = ("Dataset {name} downloaded and prepared to {data_dir}. "
                     "Subsequent calls will reuse this data.").format(
                             name=self.name,
                             data_dir=self._data_dir,
                     )
        print(msg)

    def _download_and_prepare(self, dl_manager, **prepare_split_kwargs):
        """Downloads and prepares dataset for reading.

        This is the internal implementation to overwrite called when user calls
        `download_and_prepare`. It should download all required data and generate
        the pre-processed datasets files.

        Args:
            dl_manager: (DownloadManager) `DownloadManager` used to download and cache
                data.
            download_config: `DownloadConfig`, Additional options.
        """
        os.makedirs(self._data_dir, exist_ok=True)

        # Generating data for all splits
        split_dict = splits_lib.SplitDict(dataset_name=self.name)
        split_generators_kwargs = self._make_split_generators_kwargs(
                prepare_split_kwargs)
        for split_generator in self._split_generators(
                dl_manager, **split_generators_kwargs):
            if str(split_generator.split_info.name).lower() == "all":
                raise ValueError(
                        "`all` is a special split keyword corresponding to the "
                        "union of all splits, so cannot be used as key in "
                        "._split_generator()."
                )

            logger.info("Generating split %s", split_generator.split_info.name)
            split_dict.add(split_generator.split_info)

            # Prepare split will record examples associated to the split
            self._prepare_split(split_generator, **prepare_split_kwargs)

        # Update the info object with the splits.
        self.info.update_splits_if_different(split_dict)

    def _make_download_manager(self, download_dir, download_config):
        """Creates a new download manager object."""
        download_dir = download_dir or os.path.join(self._data_dir_root, "downloads")
        extract_dir = (download_config.extract_dir or
                                     os.path.join(download_dir, "extracted"))

        # Use manual_dir only if MANUAL_DOWNLOAD_INSTRUCTIONS are set.
        if self.MANUAL_DOWNLOAD_INSTRUCTIONS:
            manual_dir = (
                    download_config.manual_dir or os.path.join(download_dir, "manual"))
            manual_dir = os.path.join(manual_dir, self.name)
        else:
            manual_dir = None

        return DownloadManager(
                dataset_name=self.name,
                download_dir=download_dir,
                extract_dir=extract_dir,
                manual_dir=manual_dir,
                manual_dir_instructions=self.MANUAL_DOWNLOAD_INSTRUCTIONS,
                force_download=(download_config.download_mode == FORCE_REDOWNLOAD),
                force_extraction=(download_config.download_mode == FORCE_REDOWNLOAD),
                register_checksums=download_config.register_checksums,
        )

    def _make_split_generators_kwargs(self, prepare_split_kwargs):
        """Get kwargs for `self._split_generators()` from `prepare_split_kwargs`."""
        del prepare_split_kwargs
        return {}

    def as_dataset(self,
                    split=None,
                    batch_size=None,
                    as_supervised=False):
        # pylint: disable=line-too-long
        """Constructs a `Dataset`.

        Callers must pass arguments as keyword arguments.

        The output types vary depending on the parameters. Examples:

        ```python
        builder = nlp.builder('imdb_reviews')
        builder.download_and_prepare()

        # Default parameters: Returns the dict of Dataset
        ds_all_dict = builder.as_dataset()
        assert isinstance(ds_all_dict, dict)
        print(ds_all_dict.keys())  # ==> ['test', 'train', 'unsupervised']

        assert isinstance(ds_all_dict['test'], Dataset)
        # Each dataset (test, train, unsup.) consists of dictionaries
        # {'label': <tf.Tensor: .. dtype=int64, numpy=1>,
        #  'text': <tf.Tensor: .. dtype=string, numpy=b"I've watched the movie ..">}
        # {'label': <tf.Tensor: .. dtype=int64, numpy=1>,
        #  'text': <tf.Tensor: .. dtype=string, numpy=b'If you love Japanese ..'>}

        # With as_supervised: Dataset only contains (feature, label) tuples
        ds_all_supervised = builder.as_dataset(as_supervised=True)
        assert isinstance(ds_all_supervised, dict)
        print(ds_all_supervised.keys())  # ==> ['test', 'train', 'unsupervised']

        assert isinstance(ds_all_supervised['test'], Dataset)
        # Each dataset (test, train, unsup.) consists of tuples (text, label)
        # (<tf.Tensor: ... dtype=string, numpy=b"I've watched the movie ..">,
        #  <tf.Tensor: ... dtype=int64, numpy=1>)
        # (<tf.Tensor: ... dtype=string, numpy=b"If you love Japanese ..">,
        #  <tf.Tensor: ... dtype=int64, numpy=1>)

        # Same as above plus requesting a particular split
        ds_test_supervised = builder.as_dataset(as_supervised=True, split='test')
        assert isinstance(ds_test_supervised, Dataset)
        # The dataset consists of tuples (text, label)
        # (<tf.Tensor: ... dtype=string, numpy=b"I've watched the movie ..">,
        #  <tf.Tensor: ... dtype=int64, numpy=1>)
        # (<tf.Tensor: ... dtype=string, numpy=b"If you love Japanese ..">,
        #  <tf.Tensor: ... dtype=int64, numpy=1>)
        ```

        Args:
            split: `nlp.SplitBase`, which subset(s) of the data to read. If None
                (default), returns all splits in a dict
                `<key: nlp.Split, value: Dataset>`.
            batch_size: `int`, batch size. Note that variable-length features will
                be 0-padded if `batch_size` is set. Users that want more custom behavior
                should use `batch_size=None` and use the `tf.data` API to construct a
                custom pipeline. If `batch_size == -1`, will return feature
                dictionaries of the whole dataset with `tf.Tensor`s instead of a
                `Dataset`.
            as_supervised: `bool`, if `True`, the returned `Dataset`
                will have a 2-tuple structure `(input, label)` according to
                `builder.info.supervised_keys`. If `False`, the default,
                the returned `Dataset` will have a dictionary with all the
                features.

        Returns:
            `Dataset`, or if `split=None`, `dict<key: nlp.Split, value:
            Dataset>`.
        """
        # pylint: enable=line-too-long
        logger.info("Constructing Dataset for split %s, from %s",
                                 split, self._data_dir)
        if not os.path.exists(self._data_dir):
            raise AssertionError(
                    ("Dataset %s: could not find data in %s. Please make sure to call "
                     "builder.download_and_prepare(), or pass download=True to "
                     "nlp.load() before trying to access the Dataset object."
                    ) % (self.name, self._data_dir_root))

        # By default, return all splits
        if split is None:
            split = {s: s for s in self.info.splits}


        # Create a dataset for each of the given splits
        build_single_dataset = functools.partial(
                self._build_single_dataset,
                batch_size=batch_size,
                as_supervised=as_supervised,
        )
        datasets = utils.map_nested(build_single_dataset, split, map_tuple=True)
        return datasets

    def _build_single_dataset(
            self,
            split,
            batch_size,
            as_supervised):
        """as_dataset for a single split."""
        if isinstance(split, str):
            split = splits_lib.Split(split)

        # Build base dataset
        ds = self._as_dataset(
                split=split,
        )
        return ds

    def _as_dataset(
            self,
            split=splits_lib.Split.TRAIN):
        """Constructs a `Dataset`.

        This is the internal implementation to overwrite called when user calls
        `as_dataset`. It should read the pre-processed datasets files and generate
        the `Dataset` object.

        Args:
            split: `nlp.Split` which subset of the data to read.

        Returns:
            `Dataset`
        """

        ds = ArrowReader(self._data_dir, self.info).read(
                name=self.name,
                instructions=split,
                split_infos=self.info.splits.values(),
        )
        return ds

    @utils.memoized_property
    def info(self):
        """`DatasetInfo` for this builder."""
        # Ensure .info hasn't been called before versioning is set-up
        # Otherwise, backward compatibility cannot be guaranteed as some code will
        # depend on the code version instead of the restored data version
        if not getattr(self, "_version", None):
            # Message for developper creating new dataset. Will trigger if they are
            # using .info in the constructor before calling super().__init__
            raise AssertionError(
                    "Info should not been called before version has been defined. "
                    "Otherwise, the created .info may not match the info version from "
                    "the restored dataset.")
        return self._info()

    @abc.abstractmethod
    def _info(self):
        """Construct the DatasetInfo object. See `DatasetInfo` for details.

        Warning: This function is only called once and the result is cached for all
        following .info() calls.

        Returns:
            info: (DatasetInfo) The dataset information
        """
        raise NotImplementedError

    @abc.abstractmethod
    def _split_generators(self, dl_manager):
        """Specify feature dictionary generators and dataset splits.

        This function returns a list of `SplitGenerator`s defining how to generate
        data and what splits to use.

        Example:

            return[
                    nlp.SplitGenerator(
                            name=nlp.Split.TRAIN,
                            gen_kwargs={'file': 'train_data.zip'},
                    ),
                    nlp.SplitGenerator(
                            name=nlp.Split.TEST,
                            gen_kwargs={'file': 'test_data.zip'},
                    ),
            ]

        The above code will first call `_generate_examples(file='train_data.zip')`
        to write the train data, then `_generate_examples(file='test_data.zip')` to
        write the test data.

        Datasets are typically split into different subsets to be used at various
        stages of training and evaluation.

        Note that for datasets without a `VALIDATION` split, you can use a
        fraction of the `TRAIN` data for evaluation as you iterate on your model
        so as not to overfit to the `TEST` data.

        For downloads and extractions, use the given `download_manager`.
        Note that the `DownloadManager` caches downloads, so it is fine to have each
        generator attempt to download the source data.

        A good practice is to download all data in this function, and then
        distribute the relevant parts to each split with the `gen_kwargs` argument

        Args:
            dl_manager: (DownloadManager) Download manager to download the data

        Returns:
            `list<SplitGenerator>`.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def _prepare_split(self, split_generator, **kwargs):
        """Generate the examples and record them on disk.

        Args:
            split_generator: `SplitGenerator`, Split generator to process
            **kwargs: Additional kwargs forwarded from _download_and_prepare (ex:
                beam pipeline)
        """
        raise NotImplementedError()


class GeneratorBasedBuilder(DatasetBuilder):
    """Base class for datasets with data generation based on dict generators.

    `GeneratorBasedBuilder` is a convenience class that abstracts away much
    of the data writing and reading of `DatasetBuilder`. It expects subclasses to
    implement generators of feature dictionaries across the dataset splits
    (`_split_generators`). See the method docstrings for details.
    """

    @abc.abstractmethod
    def _generate_examples(self, **kwargs):
        """Default function generating examples for each `SplitGenerator`.

        This function preprocess the examples from the raw data to the preprocessed
        dataset files.
        This function is called once for each `SplitGenerator` defined in
        `_split_generators`. The examples yielded here will be written on
        disk.

        Args:
            **kwargs: `dict`, Arguments forwarded from the SplitGenerator.gen_kwargs

        Yields:
            key: `str` or `int`, a unique deterministic example identification key.
                * Unique: An error will be raised if two examples are yield with the
                    same key.
                * Deterministic: When generating the dataset twice, the same example
                    should have the same key.
                Good keys can be the image id, or line number if examples are extracted
                from a text file.
                The key will be hashed and sorted to shuffle examples deterministically,
                such as generating the dataset multiple times keep examples in the
                same order.
            example: `dict<str feature_name, feature_value>`, a feature dictionary
                ready to be encoded and written to disk. The example will be
                encoded with `self.info.features.encode_example({...})`.
        """
        raise NotImplementedError()

    def _download_and_prepare(self, dl_manager, download_config):
        # Extract max_examples_per_split and forward it to _prepare_split
        super(GeneratorBasedBuilder, self)._download_and_prepare(
                dl_manager=dl_manager,
                max_examples_per_split=download_config.max_examples_per_split,
        )

    def _prepare_split(self, split_generator, max_examples_per_split):
        generator = self._generate_examples(**split_generator.gen_kwargs)
        split_info = split_generator.split_info
        if max_examples_per_split is not None:
            logger.warning("Splits capped at %s examples max.",
                                            max_examples_per_split)
            generator = itertools.islice(generator, max_examples_per_split)
        fname = "{}-{}.arrow".format(self.name, split_generator.name)
        fpath = os.path.join(self._data_dir, fname)
        examples_type = self.info.features.get_type()
        writer = ArrowWriter(data_type=examples_type, path=fpath)
        for key, record in utils.tqdm(generator, unit=" examples", total=split_info.num_examples, leave=False):
            example = self.info.features.encode_example(record)
            writer.write(example)
        num_examples, num_bytes = writer.finalize()
        assert num_examples == num_examples, f"Expected to write {split_info.num_examples} but wrote {num_examples}"
        split_generator.split_info.num_examples = num_examples
        split_generator.split_info.num_bytes = num_bytes


class BeamBasedBuilder(DatasetBuilder):
    """Beam based Builder."""

    def __init__(self, *args, **kwargs):
        super(BeamBasedBuilder, self).__init__(*args, **kwargs)
        self._beam_writers = {}  # {split: beam_writer} mapping.

    def _make_split_generators_kwargs(self, prepare_split_kwargs):
        # Pass `pipeline` into `_split_generators()` from `prepare_split_kwargs` if
        # it's in the call signature of `_split_generators()`.
        # This allows for global preprocessing in beam.
        split_generators_kwargs = {}
        split_generators_arg_names = (
                inspect.signature(self._split_generators).parameters.keys())
        if "pipeline" in split_generators_arg_names:
            split_generators_kwargs["pipeline"] = prepare_split_kwargs["pipeline"]
        return split_generators_kwargs

    @abc.abstractmethod
    def _build_pcollection(self, pipeline, **kwargs):
        """Build the beam pipeline examples for each `SplitGenerator`.

        This function extracts examples from the raw data with parallel transforms
        in a Beam pipeline. It is called once for each `SplitGenerator` defined in
        `_split_generators`. The examples from the PCollection will be
        encoded and written to disk.

        Warning: When running in a distributed setup, make sure that the data
        which will be read (download_dir, manual_dir,...) and written (data_dir)
        can be accessed by the workers jobs. The data should be located in a
        shared filesystem, like GCS.

        Example:

        ```
        def _build_pcollection(pipeline, extracted_dir):
            return (
                    pipeline
                    | beam.Create(gfile.io.listdir(extracted_dir))
                    | beam.Map(_process_file)
            )
        ```

        Args:
            pipeline: `beam.Pipeline`, root Beam pipeline
            **kwargs: Arguments forwarded from the SplitGenerator.gen_kwargs

        Returns:
            pcollection: `PCollection`, an Apache Beam PCollection containing the
                example to send to `self.info.features.encode_example(...)`.
        """
        raise NotImplementedError()

    def _download_and_prepare(self, dl_manager, download_config):
        # Create the Beam pipeline and forward it to _prepare_split
        beam = lazy_imports.apache_beam

        if not download_config.beam_runner and not download_config.beam_options:
            raise ValueError(
                    "Trying to generate a dataset using Apache Beam, yet no Beam Runner "
                    "or PipelineOptions() has been provided. Please pass a "
                    "nlp.download.DownloadConfig(beam_runner=...) object to the "
                    "builder.download_and_prepare(download_config=...) method"
            )

        beam_options = (download_config.beam_options or
                                        beam.options.pipeline_options.PipelineOptions())
        # Beam type checking assumes transforms multiple outputs are of same type,
        # which is not our case. Plus it doesn't handle correctly all types, so we
        # are better without it.
        beam_options.view_as(
                beam.options.pipeline_options.TypeOptions).pipeline_type_check = False
        # Use a single pipeline for all splits
        with beam.Pipeline(
                runner=download_config.beam_runner,
                options=beam_options,
        ) as pipeline:
            # TODO(nlp): Should eventually try to add support to
            # download_config.max_examples_per_split
            super(BeamBasedBuilder, self)._download_and_prepare(
                    dl_manager,
                    pipeline=pipeline,
            )

        # Update `info.splits` with number of shards and shard lengths.
        split_dict = self.info.splits
        for split_name, beam_writer in self._beam_writers.items():
            logger.info("Retrieving shard lengths for %s...", split_name)
            shard_lengths, total_size = beam_writer.finalize()
            split_info = split_dict[split_name]
            split_info.shard_lengths.extend(shard_lengths)
            split_info.num_shards = len(shard_lengths)
            split_info.num_bytes = total_size
        logger.info("Updating split info...")
        self.info.update_splits_if_different(split_dict)

    def _prepare_split(self, split_generator, pipeline):
        beam = lazy_imports.apache_beam

        os.makedirs(self._data_dir, exist_ok=True)

        split_name = split_generator.split_info.name
        output_prefix = filename_prefix_for_split(
                self.name, split_name)
        output_prefix = os.path.join(self._data_dir, output_prefix)

        # To write examples to disk:
        fname = "{}-{}.arrow".format(self.name, split_name)
        fpath = os.path.join(self._data_dir, fname)
        examples_type = self.info.features.get_type()
        beam_writer = BeamWriter(
                examples_type, fpath, hash_salt=split_name)
        self._beam_writers[split_name] = beam_writer

        encode_example = self.info.features.encode_example

        # Note: We need to wrap the pipeline in a PTransform to avoid re-using the
        # same label names for each split
        @beam.ptransform_fn
        def _build_pcollection(pipeline):
            """PTransformation which build a single split."""
            # Encode the PCollection
            pcoll_examples = self._build_pcollection(
                    pipeline, **split_generator.gen_kwargs)
            pcoll_examples |= "Encode" >> beam.Map(
                    lambda key_ex: (key_ex[0], encode_example(key_ex[1])))
            return beam_writer.write_from_pcollection(pcoll_examples)

        # Add the PCollection to the pipeline
        _ = pipeline | split_name >> _build_pcollection()   # pylint: disable=no-value-for-parameter



# coding=utf-8
# Copyright 2020 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Async download API with checksum verification. No business logic."""

import concurrent.futures
import contextlib
import hashlib
import io
import os
import shutil
import re
import ssl
import sys
import logging
import promise
import requests
from requests.utils import extract_zipped_paths
import urllib

from .. import utils
from . import kaggle
from . import util as download_util

_DRIVE_URL = re.compile(r'^https://drive\.google\.com/')

logger = logging.getLogger(__name__)

@utils.memoize()
def get_downloader(*args, **kwargs):
    return _Downloader(*args, **kwargs)


def _get_filename(response):
    content_disposition = response.headers.get('content-disposition', None)
    if content_disposition:
        match = re.findall('filename="(.+?)"', content_disposition)
        if match:
            return match[0]
    return download_util.get_file_name(response.url)


class DownloadError(Exception):
    pass


class _Downloader(object):
    """Class providing async download API with checksum validation.

    Do not instantiate this class directly. Instead, call `get_downloader()`.
    """

    def __init__(self, max_simultaneous_downloads=50, checksumer=None):
        """Init _Downloader instance.

        Args:
            max_simultaneous_downloads: `int`, max number of simultaneous downloads.
            checksumer: `hashlib.HASH`. Defaults to `hashlib.sha256`.
        """
        self._executor = concurrent.futures.ThreadPoolExecutor(
                max_workers=max_simultaneous_downloads)
        self._checksumer = checksumer or hashlib.sha256
        self._pbar_url = None
        self._pbar_dl_size = None

    @utils.memoize()
    def kaggle_downloader(self, competition_name):
        return kaggle.KaggleCompetitionDownloader(competition_name)

    @contextlib.contextmanager
    def tqdm(self):
        """Add a progression bar for the current download."""
        async_tqdm = utils.async_tqdm
        with async_tqdm(total=0, desc='Dl Completed...', unit=' url') as pbar_url:
            with async_tqdm(total=0, desc='Dl Size...', unit=' MiB') as pbar_dl_size:
                self._pbar_url = pbar_url
                self._pbar_dl_size = pbar_dl_size
                yield

    def download(self, url, destination_path):
        """Download url to given path.

        Returns Promise -> sha256 of downloaded file.

        Args:
            url: address of resource to download.
            destination_path: `str`, path to directory where to download the resource.

        Returns:
            Promise obj -> (`str`, int): (downloaded object checksum, size in bytes).
        """
        self._pbar_url.update_total(1)
        future = self._executor.submit(self._sync_download, url, destination_path)
        return promise.Promise.resolve(future)

    def _sync_kaggle_download(self, kaggle_url, destination_path):
        """Download with Kaggle API."""
        kaggle_file = kaggle.KaggleFile.from_url(kaggle_url)
        downloader = self.kaggle_downloader(kaggle_file.competition)
        filepath = downloader.download_file(kaggle_file.filename, destination_path)

        # dl_size = tf . io . gfile . stat(filepath).length  # TODO thom deprecating this
        checksum = self._checksumer()
        with open(filepath, 'rb') as f:
            while True:
                block = f.read(io.DEFAULT_BUFFER_SIZE)
                if not block:
                    break
                checksum.update(block)
        return checksum.hexdigest(), 0  # dl_size

    def _get_drive_url(self, url, session):
        """Returns url, possibly with confirmation token."""
        response = session.get(url, stream=True)
        if response.status_code != 200:
            raise DownloadError(
                    'Failed to get url %s. HTTP code: %d.' % (url, response.status_code))
        for k, v in response.cookies.items():
            if k.startswith('download_warning'):
                return url + '&confirm=' + v  # v is the confirm token
        # No token found, let's try with original URL:
        return url

    def _sync_file_copy(self, filepath, destination_path):
        out_path = os.path.join(destination_path, os.path.basename(filepath))
        shutil.copyfile(filepath, out_path)
        hexdigest, size = utils.read_checksum_digest(
                out_path, checksum_cls=self._checksumer)
        return hexdigest, size

    def _sync_download(self, url, destination_path):
        """Synchronous version of `download` method.

        Args:
            url: url to download
            destination_path: path where to write it
        Returns:
            None

        Raises:
            DownloadError: when download fails.

        Environment Variables:
                TFDS_HTTP_PROXY  : Configure Proxy Servers for HTTP Requests
                TFDS_HTTPS_PROXY : Configure Proxy Servers for HTTPS Requests
                TFDS_FTP_PROXY   : Configure Proxy Servers for FTP Requests
                TFDS_CA_BUNDLE   : Configure Custom Client Side SSL Certificates
                                                     (If REQUESTS_CA_BUNDLE is set, it would use that
                                                     path)

            Note: FTPS Custom Certificate verification doesn't work for python version
            <= 2.7.8
        """
        proxies = {
                'http': os.environ.get('TFDS_HTTP_PROXY', None),
                'https': os.environ.get('TFDS_HTTPS_PROXY', None),
                'ftp': os.environ.get('TFDS_FTP_PROXY', None)
        }
        ca_bundle = os.environ.get('TFDS_CA_BUNDLE', None)
        if not ca_bundle:
            ca_bundle = os.environ.get('REQUESTS_CA_BUNDLE', None)
        if ca_bundle:
            ca_bundle = extract_zipped_paths(ca_bundle)

        ca_verify = {
                'urllib':
                        ssl._create_unverified_context()  # pylint: disable=W0212
                        if not ca_bundle else ssl.create_default_context(capath=ca_bundle),
                'requests':
                        False if not ca_bundle else ca_bundle
        }

        if kaggle.KaggleFile.is_kaggle_url(url):
            if proxies['http']:
                os.environ['KAGGLE_PROXY'] = proxies['http']
            return self._sync_kaggle_download(url, destination_path)

        # If url is on a filesystem that gfile understands, use copy. Otherwise,
        # use requests (http) or urllib (ftp).
        if not url.startswith('http'):
            return self._sync_file_copy(url, destination_path)

        session = requests.Session()
        session.proxies = proxies
        session.verify = ca_verify['requests']
        if _DRIVE_URL.match(url):
            url = self._get_drive_url(url, session)
        use_urllib = url.startswith('ftp')
        if use_urllib:
            if proxies['ftp']:
                proxy = urllib.request.ProxyHandler({'ftp': proxies['ftp']})
                opener = urllib.request.build_opener(proxy)
                urllib.request.install_opener(opener)   # pylint: disable=too-many-function-args
            request = urllib.request.Request(url)

            # disable ssl context check for FTPS for python version <= 2.7.8
            if ca_verify['urllib'] is None:
                response = urllib.request.urlopen(request)
            else:
                response = urllib.request.urlopen(request, context=ca_verify['urllib'])
        else:
            response = session.get(url, stream=True)
            if response.status_code != 200:
                raise DownloadError('Failed to get url %s. HTTP code: %d.' %
                                                        (url, response.status_code))
        fname = _get_filename(response)
        path = os.path.join(destination_path, fname)
        size = 0

        size_mb = 0
        unit_mb = 2**20  # MB
        self._pbar_dl_size.update_total(
                int(response.headers.get('Content-length', 0)) // unit_mb)
        with open(path, 'wb') as file_:
            checksum = self._checksumer()
            if use_urllib:
                iterator = iter(lambda: response.read(io.DEFAULT_BUFFER_SIZE), b'')
            else:
                iterator = response.iter_content(chunk_size=io.DEFAULT_BUFFER_SIZE)

            for block in iterator:
                size += len(block)

                # Update the progress bar
                size_mb += len(block)
                if size_mb > unit_mb:
                    self._pbar_dl_size.update(size_mb // unit_mb)
                    size_mb %= unit_mb

                checksum.update(block)
                file_.write(block)
        self._pbar_url.update(1)
        return checksum.hexdigest(), size



# coding=utf-8
# Copyright 2020 The HuggingFace NLP Authors and the TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
# pylint: disable=line-too-long
"""`nlp` defines a collection of datasets ready-to-use.

Each dataset is defined as a `nlp.DatasetBuilder`, which encapsulates
the logic to download the dataset and construct an input pipeline, as well as
contains the dataset documentation (version, splits, number of examples, etc.).

The main library entrypoints are:

* `nlp.builder`: fetch a `nlp.DatasetBuilder` by name
* `nlp.load`: convenience method to construct a builder, download the data, and
    create an input pipeline, returning an NLP dataset.

"""
# pylint: enable=line-too-long
# pylint: disable=g-import-not-at-top,g-bad-import-order,wrong-import-position

__version__ = "0.0.1"

# Types are pyarrow types
from pyarrow import (null, bool_,
                     int8, int16, int32, int64,
                     uint8, uint16, uint32, uint64,
                     time32, time64, timestamp, date32, date64, duration,
                     float16, float32, float64,
                     binary, string, utf8,
                     large_binary, large_string, large_utf8,
                     decimal128,
                     list_, large_list, map_, struct, union, dictionary)
from pyarrow import total_allocated_bytes

from . import download, features, load, datasets
from .load import builder, load
from .download import GenerateMode
from .arrow_dataset import Dataset
from .splits import percent
from .splits import Split
from .utils.tqdm_utils import disable_progress_bar

from .builder import BeamBasedBuilder, BuilderConfig, DatasetBuilder, GeneratorBasedBuilder

from .info import DatasetInfo

from .lazy_imports_lib import lazy_imports

from .splits import NamedSplit, SplitBase, SplitDict, SplitGenerator, SplitInfo, SubSplitInfo

from .arrow_reader import ReadInstruction

from .utils import *



# coding=utf-8
# Copyright 2020 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Download manager interface."""

import os
import shutil
import sys
import uuid

import logging
import promise

from .. import utils
from . import checksums
from . import downloader
from . import extractor
from . import resource as resource_lib
from . import util

logger = logging.getLogger(__name__)

class NonMatchingChecksumError(Exception):
    """The downloaded file doesn't have expected checksum."""

    def __init__(self, url, tmp_path):
        msg = 'Artifact %s, downloaded to %s, has wrong checksum.' % (url, tmp_path)
        Exception.__init__(self, msg)


class DownloadConfig(object):
    """Configuration for `nlp.DatasetBuilder.download_and_prepare`."""

    def __init__(self,
                             extract_dir=None,
                             manual_dir=None,
                             download_mode=None,
                             compute_stats=None,
                             max_examples_per_split=None,
                             register_checksums=False,
                             beam_runner=None,
                             beam_options=None):
        """Constructs a `DownloadConfig`.

        Args:
            extract_dir: `str`, directory where extracted files are stored.
                Defaults to "<download_dir>/extracted".
            manual_dir: `str`, read-only directory where manually downloaded/extracted
                data is stored. Defaults to
                "<download_dir>/manual".
            download_mode: `nlp.GenerateMode`, how to deal with downloads or data
                that already exists. Defaults to `REUSE_DATASET_IF_EXISTS`, which will
                reuse both downloads and data if it already exists.
            compute_stats: `nlp.download.ComputeStats`, whether to compute
                statistics over the generated data. Defaults to `AUTO`.
            max_examples_per_split: `int`, optional max number of examples to write
                into each split (used for testing).
            register_checksums: `bool`, defaults to False. If True, checksum of
                downloaded files are recorded.
            beam_runner: Runner to pass to `beam.Pipeline`, only used for datasets
                based on Beam for the generation.
            beam_options: `PipelineOptions` to pass to `beam.Pipeline`, only used for
                datasets based on Beam for the generation.
        """
        self.extract_dir = extract_dir
        self.manual_dir = manual_dir
        self.download_mode = util.GenerateMode(
                download_mode or util.GenerateMode.REUSE_DATASET_IF_EXISTS)
        self.compute_stats = util.ComputeStatsMode(
                compute_stats or util.ComputeStatsMode.AUTO)
        self.max_examples_per_split = max_examples_per_split
        self.register_checksums = register_checksums
        self.beam_runner = beam_runner
        self.beam_options = beam_options


class DownloadManager(object):
    """Manages the download and extraction of files, as well as caching.

    Downloaded files are cached under `download_dir`. The file name of downloaded
     files follows pattern "{sanitized_url}{content_checksum}.{ext}". Eg:
     'cs.toronto.edu_kriz_cifar-100-pythonJDF[...]I.tar.gz'.

    While a file is being downloaded, it is placed into a directory following a
    similar but different pattern:
    "{sanitized_url}{url_checksum}.tmp.{uuid}".

    When a file is downloaded, a "{fname}.INFO.json" file is created next to it.
    This INFO file contains the following information:
    {"dataset_names": ["name1", "name2"],
     "urls": ["http://url.of/downloaded_file"]}

    Extracted files/dirs are stored under `extract_dir`. The file name or
    directory name is the same as the original name, prefixed with the extraction
    method. E.g.
     "{extract_dir}/TAR_GZ.cs.toronto.edu_kriz_cifar-100-pythonJDF[...]I.tar.gz".

    The function members accept either plain value, or values wrapped into list
    or dict. Giving a data structure will parallelize the downloads.

    Example of usage:

    ```
    # Sequential download: str -> str
    train_dir = dl_manager.download_and_extract('https://abc.org/train.tar.gz')
    test_dir = dl_manager.download_and_extract('https://abc.org/test.tar.gz')

    # Parallel download: list -> list
    image_files = dl_manager.download(
            ['https://a.org/1.jpg', 'https://a.org/2.jpg', ...])

    # Parallel download: dict -> dict
    data_dirs = dl_manager.download_and_extract({
         'train': 'https://abc.org/train.zip',
         'test': 'https://abc.org/test.zip',
    })
    data_dirs['train']
    data_dirs['test']
    ```

    For more customization on the download/extraction (ex: passwords, output_name,
    ...), you can pass a `nlp.download.Resource` as argument.
    """

    def __init__(self,
                             download_dir,
                             extract_dir=None,
                             manual_dir=None,
                             manual_dir_instructions=None,
                             dataset_name=None,
                             force_download=False,
                             force_extraction=False,
                             register_checksums=False):
        """Download manager constructor.

        Args:
            download_dir: `str`, path to directory where downloads are stored.
            extract_dir: `str`, path to directory where artifacts are extracted.
            manual_dir: `str`, path to manually downloaded/extracted data directory.
            manual_dir_instructions: `str`, human readable instructions on how to
                                                 prepare contents of the manual_dir for this dataset.
            dataset_name: `str`, name of dataset this instance will be used for. If
                provided, downloads will contain which datasets they were used for.
            force_download: `bool`, default to False. If True, always [re]download.
            force_extraction: `bool`, default to False. If True, always [re]extract.
            register_checksums: `bool`, default to False. If True, dl checksums aren't
                checked, but stored into file.
        """
        self._dataset_name = dataset_name
        self._download_dir = os.path.expanduser(download_dir)
        self._extract_dir = os.path.expanduser(
                extract_dir or os.path.join(download_dir, 'extracted'))
        self._manual_dir = manual_dir and os.path.expanduser(manual_dir)
        self._manual_dir_instructions = manual_dir_instructions
        os.makedirs(self._download_dir, exist_ok=True)
        os.makedirs(self._extract_dir, exist_ok=True)
        self._force_download = force_download
        self._force_extraction = force_extraction
        self._extractor = extractor.get_extractor()
        self._downloader = downloader.get_downloader()
        self._register_checksums = register_checksums
        # All known URLs: {url: (size, checksum)}
        self._sizes_checksums = {}  # checksums.get_all_sizes_checksums()
        # To record what is being used: {url: (size, checksum)}
        self._recorded_sizes_checksums = {}

    @property
    def downloaded_size(self):
        """Returns the total size of downloaded files."""
        return sum(size for size, sha256 in self._recorded_sizes_checksums.values())

    def _get_final_dl_path(self, url, sha256):
        return os.path.join(self._download_dir,
                                                resource_lib.get_dl_fname(url, sha256))

    @util.build_synchronize_decorator()
    def _record_sizes_checksums(self):
        """Store in file when recorded size/checksum of downloaded files."""
        checksums.store_checksums(self._dataset_name,self._recorded_sizes_checksums)

    def _handle_download_result(self, resource, tmp_dir_path, sha256, dl_size):
        """Store dled file to definitive place, write INFO file, return path."""
        fnames = os.listdir(tmp_dir_path)
        if len(fnames) > 1:
            raise AssertionError('More than one file in %s.' % tmp_dir_path)
        original_fname = fnames[0]
        tmp_path = os.path.join(tmp_dir_path, original_fname)
        self._recorded_sizes_checksums[resource.url] = (dl_size, sha256)
        if self._register_checksums:
            self._record_sizes_checksums()
        # TODO: add checksum checking back
        # elif (dl_size, sha256) != self._sizes_checksums.get(resource.url, None):
        #     raise NonMatchingChecksumError(resource.url, tmp_path)
        download_path = self._get_final_dl_path(resource.url, sha256)
        resource_lib.write_info_file(resource, download_path, self._dataset_name,
                                                                 original_fname)
        # Unconditionally overwrite because either file doesn't exist or
        # FORCE_DOWNLOAD=true
        os.rename(tmp_path, download_path)
        shutil.rmtree(tmp_dir_path)
        return download_path

    def download_checksums(self, checksums_url):
        """Downloads checksum file from the given URL and adds it to registry."""
        checksums_path = self.download(checksums_url)
        with open(checksums_path) as f:
            self._sizes_checksums.update(checksums.parse_sizes_checksums(f))

    # synchronize and memoize decorators ensure same resource will only be
    # processed once, even if passed twice to download_manager.
    @util.build_synchronize_decorator()
    @utils.memoize()
    def _download(self, resource):
        """Download resource, returns Promise->path to downloaded file."""
        if isinstance(resource, str):
            resource = resource_lib.Resource(url=resource)
        url = resource.url
        if url in self._sizes_checksums:
            expected_sha256 = self._sizes_checksums[url][1]
            download_path = self._get_final_dl_path(url, expected_sha256)
            if not self._force_download and resource.exists_locally(download_path):
                logger.info('URL %s already downloaded: reusing %s.',
                                         url, download_path)
                self._recorded_sizes_checksums[url] = self._sizes_checksums[url]
                return promise.Promise.resolve(download_path)
        # There is a slight difference between downloader and extractor here:
        # the extractor manages its own temp directory, while the DownloadManager
        # manages the temp directory of downloader.
        download_dir_path = os.path.join(
                self._download_dir,
                '%s.tmp.%s' % (resource_lib.get_dl_dirname(url), uuid.uuid4().hex))
        os.makedirs(download_dir_path)
        logger.info('Downloading %s into %s...', url, download_dir_path)

        def callback(val):
            checksum, dl_size = val
            return self._handle_download_result(
                    resource, download_dir_path, checksum, dl_size)
        return self._downloader.download(url, download_dir_path).then(callback)

    @util.build_synchronize_decorator()
    @utils.memoize()
    def _extract(self, resource):
        """Extract a single archive, returns Promise->path to extraction result."""
        if isinstance(resource, str):
            resource = resource_lib.Resource(path=resource)
        path = resource.path
        extract_method = resource.extract_method
        if extract_method == resource_lib.ExtractMethod.NO_EXTRACT:
            logger.info('Skipping extraction for %s (method=NO_EXTRACT).', path)
            return promise.Promise.resolve(path)
        method_name = resource_lib.ExtractMethod(extract_method).name
        extract_path = os.path.join(self._extract_dir,
                                                                '%s.%s' % (method_name, os.path.basename(path)))
        if not self._force_extraction and os.path.exists(extract_path):
            logger.info('Reusing extraction of %s at %s.', path, extract_path)
            return promise.Promise.resolve(extract_path)
        return self._extractor.extract(path, extract_method, extract_path)

    @util.build_synchronize_decorator()
    @utils.memoize()
    def _download_extract(self, resource):
        """Download-extract `Resource` or url, returns Promise->path."""
        if isinstance(resource, str):
            resource = resource_lib.Resource(url=resource)
        def callback(path):
            resource.path = path
            return self._extract(resource)
        return self._download(resource).then(callback)

    def download_kaggle_data(self, competition_name):
        """Download data for a given Kaggle competition."""
        with self._downloader.tqdm():
            kaggle_downloader = self._downloader.kaggle_downloader(competition_name)
            urls = kaggle_downloader.competition_urls
            files = kaggle_downloader.competition_files
            return _map_promise(self._download,
                                                    dict((f, u) for (f, u) in zip(files, urls)))

    def download(self, url_or_urls):
        """Download given url(s).

        Args:
            url_or_urls: url or `list`/`dict` of urls to download and extract. Each
                url can be a `str` or `nlp.download.Resource`.

        Returns:
            downloaded_path(s): `str`, The downloaded paths matching the given input
                url_or_urls.
        """
        # Add progress bar to follow the download state
        with self._downloader.tqdm():
            return _map_promise(self._download, url_or_urls)

    def iter_archive(self, resource):
        """Returns iterator over files within archive.

        **Important Note**: caller should read files as they are yielded.
        Reading out of order is slow.

        Args:
            resource: path to archive or `nlp.download.Resource`.

        Returns:
            Generator yielding tuple (path_within_archive, file_obj).
        """
        if isinstance(resource, str):
            resource = resource_lib.Resource(path=resource)
        return extractor.iter_archive(resource.path, resource.extract_method)

    def extract(self, path_or_paths):
        """Extract given path(s).

        Args:
            path_or_paths: path or `list`/`dict` of path of file to extract. Each
                path can be a `str` or `nlp.download.Resource`.

        If not explicitly specified in `Resource`, the extraction method is deduced
        from downloaded file name.

        Returns:
            extracted_path(s): `str`, The extracted paths matching the given input
                path_or_paths.
        """
        # Add progress bar to follow the download state
        with self._extractor.tqdm():
            return _map_promise(self._extract, path_or_paths)

    def download_and_extract(self, url_or_urls):
        """Download and extract given url_or_urls.

        Is roughly equivalent to:

        ```
        extracted_paths = dl_manager.extract(dl_manager.download(url_or_urls))
        ```

        Args:
            url_or_urls: url or `list`/`dict` of urls to download and extract. Each
                url can be a `str` or `nlp.download.Resource`.

        If not explicitly specified in `Resource`, the extraction method will
        automatically be deduced from downloaded file name.

        Returns:
            extracted_path(s): `str`, extracted paths of given URL(s).
        """
        # Add progress bar to follow the download state
        with self._downloader.tqdm():
            with self._extractor.tqdm():
                return _map_promise(self._download_extract, url_or_urls)

    @property
    def manual_dir(self):
        """Returns the directory containing the manually extracted data."""
        if not self._manual_dir:
            raise AssertionError(
                    'Manual directory was enabled. '
                    'Did you set MANUAL_DOWNLOAD_INSTRUCTIONS in your dataset?')
        if not os.path.exists(self._manual_dir):
            raise AssertionError(
                    'Manual directory {} does not exist. Create it and download/extract '
                    'dataset artifacts in there. Additional instructions: {}'.format(
                            self._manual_dir, self._manual_dir_instructions))
        return self._manual_dir


# ============================================================================
# In Python 2.X, threading.Condition.wait() cannot be interrupted by SIGINT,
# unless it's given a timeout. Here we artificially give a long timeout to
# allow ctrl+C.
# This code should be deleted once python2 is no longer supported.
if sys.version_info[0] > 2:

    def _wait_on_promise(p):
        return p.get()

else:

    def _wait_on_promise(p):
        while True:
            result = p.get(sys.maxint)  # pylint: disable=g-deprecated-member-used
            if p.is_fulfilled:
                return result

# ============================================================================


def _map_promise(map_fn, all_inputs):
    """Map the function into each element and resolve the promise."""
    all_promises = utils.map_nested(map_fn, all_inputs)  # Apply the function
    res = utils.map_nested(_wait_on_promise, all_promises)
    return res



# coding=utf-8
# Copyright 2020 The TensorFlow Datasets Authors and the HuggingFace NLP Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""SQUAD: The Stanford Question Answering Dataset."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import json
import os

import logging
import nlp

_CITATION = """\
@article{2016arXiv160605250R,
             author = {{Rajpurkar}, Pranav and {Zhang}, Jian and {Lopyrev},
                                 Konstantin and {Liang}, Percy},
                title = "{SQuAD: 100,000+ Questions for Machine Comprehension of Text}",
            journal = {arXiv e-prints},
                 year = 2016,
                    eid = {arXiv:1606.05250},
                pages = {arXiv:1606.05250},
archivePrefix = {arXiv},
             eprint = {1606.05250},
}
"""

_DESCRIPTION = """\
Stanford Question Answering Dataset (SQuAD) is a reading comprehension \
dataset, consisting of questions posed by crowdworkers on a set of Wikipedia \
articles, where the answer to every question is a segment of text, or span, \
from the corresponding reading passage, or the question might be unanswerable.
"""


class SquadConfig(nlp.BuilderConfig):
    """BuilderConfig for SQUAD."""

    def __init__(self, **kwargs):
        """BuilderConfig for SQUAD.

        Args:
            **kwargs: keyword arguments forwarded to super.
        """
        super(SquadConfig, self).__init__(**kwargs)


class Squad(nlp.GeneratorBasedBuilder):
    """SQUAD: The Stanford Question Answering Dataset. Version 1.1."""
    _URL = "https://rajpurkar.github.io/SQuAD-explorer/dataset/"
    _DEV_FILE = "dev-v1.1.json"
    _TRAINING_FILE = "train-v1.1.json"

    BUILDER_CONFIGS = [
            SquadConfig(
                    name="plain_text",
                    version=nlp.Version(
                            "1.0.0",
                            "New split API (https://tensorflow.org/datasets/splits)"),
                    description="Plain text",
            ),
    ]

    def _info(self):
        return nlp.DatasetInfo(
                builder=self,
                description=_DESCRIPTION,
                features=nlp.features.FeaturesDict({
                        "id":
                                nlp.string,
                        "title":
                                nlp.string,
                        "context":
                                nlp.string,
                        "question":
                                nlp.string,
                        "answers":
                                nlp.features.Sequence({
                                        "text": nlp.string,
                                        "answer_start": nlp.int32,
                                }),
                }),
                # No default supervised_keys (as we have to pass both question
                # and context as input).
                supervised_keys=None,
                homepage="https://rajpurkar.github.io/SQuAD-explorer/",
                citation=_CITATION,
        )

    def _split_generators(self, dl_manager):
        urls_to_download = {
                "train": os.path.join(self._URL, self._TRAINING_FILE),
                "dev": os.path.join(self._URL, self._DEV_FILE)
        }
        downloaded_files = dl_manager.download_and_extract(urls_to_download)

        return [
                nlp.SplitGenerator(
                        name=nlp.Split.TRAIN,
                        gen_kwargs={"filepath": downloaded_files["train"]}),
                nlp.SplitGenerator(
                        name=nlp.Split.VALIDATION,
                        gen_kwargs={"filepath": downloaded_files["dev"]}),
        ]

    def _generate_examples(self, filepath):
        """This function returns the examples in the raw (text) form."""
        logging.info("generating examples from = %s", filepath)
        with open(filepath) as f:
            squad = json.load(f)
            for article in squad["data"]:
                title = article.get("title", "").strip()
                for paragraph in article["paragraphs"]:
                    context = paragraph["context"].strip()
                    for qa in paragraph["qas"]:
                        question = qa["question"].strip()
                        id_ = qa["id"]

                        answer_starts = [answer["answer_start"] for answer in qa["answers"]]
                        answers = [answer["text"].strip() for answer in qa["answers"]]

                        # Features currently used are "context", "question", and "answers".
                        # Others are extracted here for the ease of future expansions.
                        yield id_, {
                                "title": title,
                                "context": context,
                                "question": question,
                                "id": id_,
                                "answers": {
                                        "answer_start": answer_starts,
                                        "text": answers,
                                },
                        }



# coding=utf-8
# Copyright 2020 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Methods to retrieve and store size/checksums associated to URLs.

"""

import os

from .. import utils


_ROOT_DIR = os.path.normpath(os.path.join(os.path.dirname(__file__), '../..'))

_CHECKSUM_DIRS = [
		os.path.join(_ROOT_DIR, 'url_checksums'),
]
_CHECKSUM_SUFFIX = '.txt'


def add_checksums_dir(checksums_dir):
	"""Registers a new checksums dir.

	This function allow external datasets not present in the nlp repository to
	define their own checksums_dir containing the dataset downloads checksums.

	Note: When redistributing your dataset, you should distribute the checksums
	files with it and set `add_checksums_dir` when the user is importing your
	`my_dataset.py`.

	```
	# Set-up the folder containing the 'my_dataset.txt' checksums.
	checksum_dir = os.path.join(os.path.dirname(__file__), 'checksums/')
	checksum_dir = os.path.normpath(checksum_dir)

	# Add the checksum dir (will be executed when the user import your dataset)
	nlp.download.add_checksums_dir(checksum_dir)

	class MyDataset(nlp.DatasetBuilder):
		...
	```

	Args:
		checksums_dir: `str`, checksums dir to add to the registry
	"""
	if checksums_dir in _CHECKSUM_DIRS:  # Avoid duplicate
		return
	_CHECKSUM_DIRS.append(checksums_dir)


def _list_dir(path):
	return os.listdir(path)




@utils.memoize()
def _checksum_paths():
	"""Returns dict {'dataset_name': 'path/to/checksums/file'}."""
	dataset2path = {}
	for dir_path in _CHECKSUM_DIRS:
		for fname in _list_dir(dir_path):
			if not fname.endswith(_CHECKSUM_SUFFIX):
				continue
			fpath = os.path.join(dir_path, fname)
			dataset_name = fname[:-len(_CHECKSUM_SUFFIX)]
			dataset2path[dataset_name] = fpath
	return dataset2path


def _get_path(dataset_name):
	"""Returns path to where checksums are stored for a given dataset."""
	path = _checksum_paths().get(dataset_name, None)
	if path:
		return path
	msg = (
			'No checksums file could be find for dataset {}. Please '
			'create one in one of:\n{}'
			'If you are developing your own dataset outsite nlp, you can register '
			'your own checksums_dir with `nlp.download.add_checksums_dir('
			'checksum_dir)` or pass it to the download_and_prepare script with '
			'`--checksum_dir=`'
	).format(
			dataset_name,
			''.join(['* {}\n'.format(c) for c in _CHECKSUM_DIRS]))
	raise AssertionError(msg)


def _read_file(path):
	return open(path).read()




def _get_sizes_checksums(checksums_path):
	"""Returns {URL: (size, checksum)}s stored within file at given path."""
	checksums_file = _read_file(checksums_path).split('\n')
	return parse_sizes_checksums(checksums_file)


def parse_sizes_checksums(checksums_file):
	"""Returns {URL: (size, checksum)}s stored within given file."""
	checksums = {}
	for line in checksums_file:
		line = line.strip()  # Remove the trailing '\r' on Windows OS.
		if not line or line.startswith('#'):
			continue
		# URL might have spaces inside, but size and checksum will not.
		url, size, checksum = line.rsplit(' ', 2)
		checksums[url] = (int(size), checksum)
	return checksums


@utils.memoize()
def get_all_sizes_checksums():
	"""Returns dict associating URL to (size, sha256)."""
	sizes_checksums = {}
	for path in _checksum_paths().values():
		data = _get_sizes_checksums(path)
		for url, size_checksum in data.items():
			if (url in sizes_checksums and
					sizes_checksums[url] != size_checksum):
				raise AssertionError(
						'URL %s has 2+ distinct size/checksum tuples.' % url)
		sizes_checksums.update(data)
	return sizes_checksums


def store_checksums(dataset_name, sizes_checksums):
	"""Store given checksums and sizes for specific dataset.

	Content of file is never disgarded, only updated. This is to ensure that if
	process is killed right after first download finishes, checksums
	during previous runs aren't lost.

	It is the responsibility of the caller not to call function multiple times in
	parallel for a given dataset.

	Only original file content is updated. This means the entire set of new sizes
	and checksums must be given at every call.

	Args:
		dataset_name: string.
		sizes_checksums: dict, {url: (size_in_bytes, checksum)}.
	"""
	path = _get_path(dataset_name)
	original_data = _get_sizes_checksums(path)
	new_data = original_data.copy()
	new_data.update(sizes_checksums)
	if original_data == new_data:
		return
	with open(path, 'w') as f:
		for url, (size, checksum) in sorted(new_data.items()):
			f.write('%s %s %s\n' % (url, size, checksum))



# coding=utf-8
# Copyright 2020 The HuggingFace NLP Authors and the TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
r"""Generate the minimal source code for a new dataset.

python -m nlp.scripts.create_new_dataset \
    --dataset dataset_name \
    --type dataset_type

"""

import os

from absl import app
from absl import flags

from nlp import naming  # pylint: disable=g-import-not-at-top
from nlp import py_utils

FLAGS = flags.FLAGS

_DATASET_TYPE = [
        'audio',
        'image',
        'object_detection',
        'structured',
        'summarization',
        'text',
        'translate',
        'video',
]

flags.DEFINE_string('nlp_dir', None, 'Root directory of nlp (auto-computed)')
flags.DEFINE_string('dataset', None, 'Dataset name')
flags.DEFINE_enum('type', None, _DATASET_TYPE, 'Dataset type')


_HEADER = """\
\"""{TODO}: Add a description here.\"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

"""

_DATASET_DEFAULT_IMPORTS = """\
import nlp.public_api as nlp\n
"""

_DATASET_TEST_DEFAULTS_IMPORTS = """\
from nlp import testing
from nlp.{dataset_type} import {dataset_name}

"""

_CITATION = """\
# {TODO}: BibTeX citation
_CITATION = \"""
\"""\n
"""

_DESCRIPTION = """\
# {TODO}:
_DESCRIPTION = \"""
\"""\n
"""

_DATASET_DEFAULTS = """\

class {dataset_cls}(nlp.GeneratorBasedBuilder):
    \"""{TODO}: Short description of my dataset.\"""

    # {TODO}: Set up version.
    VERSION = nlp.Version('0.1.0')

    def _info(self):
        # {TODO}: Specifies the nlp.DatasetInfo object
        return nlp.DatasetInfo(
                builder=self,
                # This is the description that will appear on the datasets page.
                description=_DESCRIPTION,
                # nlp.features.FeatureConnectors
                features=nlp.features.FeaturesDict({{
                        # These are the features of your dataset like images, labels ...
                }}),
                # If there's a common (input, target) tuple from the features,
                # specify them here. They'll be used if as_supervised=True in
                # builder.as_dataset.
                supervised_keys=(),
                # Homepage of the dataset for documentation
                homepage='https://dataset-homepage/',
                citation=_CITATION,
        )

    def _split_generators(self, dl_manager):
        \"""Returns SplitGenerators.\"""
        # {TODO}: Downloads the data and defines the splits
        # dl_manager is a nlp.download.DownloadManager that can be used to
        # download and extract URLs
        return [
                nlp.SplitGenerator(
                        name=nlp.Split.TRAIN,
                        # These kwargs will be passed to _generate_examples
                        gen_kwargs={{}},
                ),
        ]

    def _generate_examples(self):
        \"""Yields examples.\"""
        # {TODO}: Yields (key, example) tuples from the dataset
        yield 'key', {{}}\n
"""

_DATASET_TEST_DEFAULTS = """\

class {dataset_cls}Test(testing.DatasetBuilderTestCase):
    # {TODO}:
    DATASET_CLASS = {dataset_name}.{dataset_cls}
    SPLITS = {{
            "train": 3,  # Number of fake train example
            "test": 1,  # Number of fake test example
    }}

    # If you are calling `download/download_and_extract` with a dict, like:
    #   dl_manager.download({{'some_key': 'http://a.org/out.txt', ...}})
    # then the tests needs to provide the fake output paths relative to the
    # fake data directory
    # DL_EXTRACT_RESULT = {{'some_key': 'output_file1.txt', ...}}


if __name__ == "__main__":
    testing.test_main()

"""

_CHECKSUM_FILE = """\
# {TODO}: If your dataset downloads files, then the checksums will be
# automatically added here when running the download_and_prepare script
# with --register_checksums.
"""


def create_dataset_file(root_dir, data):
    """Create a new dataset from a template."""
    file_path = os.path.join(root_dir, '{dataset_type}', '{dataset_name}.py')
    context = (
            _HEADER + _DATASET_DEFAULT_IMPORTS + _CITATION + _DESCRIPTION +
            _DATASET_DEFAULTS)

    with open(file_path.format(**data), 'w') as f:
        f.write(context.format(**data))


def add_the_init(root_dir, data):
    """Append the new dataset file to the __init__.py."""
    init_file = os.path.join(root_dir, '{dataset_type}', '__init__.py')
    context = ('from nlp.{dataset_type}.{dataset_name} import '
                         '{dataset_cls}  # {TODO} Sort alphabetically\n')
    with open(init_file.format(**data), 'a') as f:
        f.write(context.format(**data))


def create_dataset_test_file(root_dir, data):
    """Create the test file associated with the dataset."""
    file_path = os.path.join(root_dir, '{dataset_type}', '{dataset_name}_test.py')
    context = (_HEADER + _DATASET_TEST_DEFAULTS_IMPORTS + _DATASET_TEST_DEFAULTS)

    with open(file_path.format(**data), 'w') as f:
        f.write(context.format(**data))


def create_fake_data(root_dir, data):
    fake_examples_dir = os.path.join(root_dir, 'testing', 'test_data',
                                                                     'fake_examples', '{dataset_name}')
    fake_examples_dir = fake_examples_dir.format(**data)
    gfile.makedirs(fake_examples_dir)

    fake_path = os.path.join(fake_examples_dir,
                                                     'TODO-add_fake_data_in_this_directory.txt')
    with open(fake_path, 'w') as f:
        f.write('{TODO}: Add fake data in this directory'.format(**data))


def create_checksum_file(root_dir, data):
    checksum_path = os.path.join(root_dir, 'url_checksums', '{dataset_name}.txt')
    with open(checksum_path.format(**data), 'w') as f:
        f.write(_CHECKSUM_FILE.format(**data))


def main(_):
    dataset_name = FLAGS.dataset
    dataset_type = FLAGS.type
    root_dir = FLAGS.nlp_dir
    if not root_dir:
        root_dir = py_utils.nlp_dir()

    data = dict(
            dataset_name=dataset_name,
            dataset_type=dataset_type,
            dataset_cls=naming.snake_to_camelcase(dataset_name),
            TODO='TODO({})'.format(dataset_name),
    )

    create_dataset_file(root_dir, data)
    add_the_init(root_dir, data)
    create_dataset_test_file(root_dir, data)
    create_fake_data(root_dir, data)
    create_checksum_file(root_dir, data)

    print(
            'Dataset generated in {}\n'
            'You can start with searching TODO({}).\n'
            'Please check this '
            '`https://github.com/tensorflow/datasets/blob/master/docs/add_dataset.md`'
            'for details.'.format(root_dir, dataset_name))


if __name__ == '__main__':
    app.run(main)



# coding=utf-8
# Copyright 2020 The HuggingFace NLP Authors and the TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
r"""Script to call download_and_prepare on DatasetBuilder.

Standalone script to generate specific dataset(s). This can be
used if you want to separate download/generation of dataset from actual usage.

By default, the dataset is generated in the default location
(~/nlp), which the same as when calling `nlp.load()`.

Instructions:
```
python -m nlp.scripts.download_and_prepare \
    --datasets=cifar10
```

If you have your dataset defined outside of `nlp`, use
`--module_import="path.to.my.dataset_module"` to have your Python module
containing your `DatasetBuilder` definition imported.


"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import importlib
import os
import pdb
import time

from absl import app
from absl import flags
import logging
import tensorflow.compat.v2 as tf
import nlp

logger = logging.getLogger(__name__)

FLAGS = flags.FLAGS

DEFAULT_DATA_DIR = os.path.expanduser(os.path.join("~", "nlp"))

flags.DEFINE_string("datasets", "",
                                        "Comma separated list of datasets to build, defaults to all"
                                        "builders.")
flags.DEFINE_string("exclude_datasets", "",
                                        "Comma separated list of datasets to exclude,"
                                        "(no download, no prepare).")
flags.DEFINE_multi_string(
        "module_import", None,
        "Modules to import. Use this when your DatasetBuilder is defined outside "
        "of nlp.")
flags.DEFINE_integer(
        "builder_config_id", None,
        "If given 1 dataset with BUILDER_CONFIGS, id of config to build.")
flags.DEFINE_boolean(
        "experimental_latest_version", False,
        "Set to true to builder the latest version available, even if not default.")

flags.DEFINE_string("data_dir", DEFAULT_DATA_DIR, "Where to place the data.")
flags.DEFINE_string("download_dir", None, "Where to place downloads.")
flags.DEFINE_string("extract_dir", None, "Where to extract files.")
flags.DEFINE_string(
        "manual_dir", None,
        "Directory where dataset have manually been downloaded / extracted.")
flags.DEFINE_string("checksums_dir", None,
                                        "For external datasets, specify the location of the "
                                        "dataset checksums.")
default_compute_stats = nlp.download.ComputeStatsMode.AUTO
flags.DEFINE_enum(
        "compute_stats",
        default_compute_stats.value,
        [e.value for e in nlp.download.ComputeStatsMode],
        "Whether to compute or not the dynamic statistics.")
flags.DEFINE_integer(
        "max_examples_per_split", None,
        "optional max number of examples to write into each split (for testing).")

# Beam flags
flags.DEFINE_list(
        "beam_pipeline_options", [],
        "A (comma-separated) list of flags to pass to `PipelineOptions` when "
        "preparing with Apache Beam. Example: "
        "`--beam_pipeline_options=job_name=my-job,project=my-project`")


# Development flags
flags.DEFINE_boolean("register_checksums", False,
                                         "If True, store size and checksum of downloaded files.")

# Debug flags
flags.DEFINE_boolean("debug", False,
                                         "If True, will drop into debugger after data generation")
flags.DEFINE_boolean("debug_start", False,
                                         "If True, will drop into debugger on startup")
flags.DEFINE_boolean("sleep_start", False,
                                         "If True, will sleep on startup; useful for ssh")
flags.DEFINE_boolean("disable_tqdm", False, "If True, disable tqdm.")


def download_config():
    return nlp.download.DownloadConfig(
            extract_dir=FLAGS.extract_dir,
            manual_dir=FLAGS.manual_dir,
            compute_stats=FLAGS.compute_stats,
            # TODO(b/116270825): Add flag to force extraction / preparation.
            download_mode=nlp.download.GenerateMode.REUSE_DATASET_IF_EXISTS,
            max_examples_per_split=FLAGS.max_examples_per_split,
            register_checksums=FLAGS.register_checksums,
    )


def download_and_prepare(builder):
    """Generate data for a given dataset."""
    logger.info("download_and_prepare for dataset %s...", builder.info.full_name)

    dl_config = download_config()

    if isinstance(builder, nlp.BeamBasedBuilder):
        beam = nlp.lazy_imports.apache_beam
        # TODO(b/129149715): Restore compute stats. Currently skipped because not
        # beam supported.
        dl_config.compute_stats = nlp.download.ComputeStatsMode.SKIP
        dl_config.beam_options = beam.options.pipeline_options.PipelineOptions(
                flags=["--%s" % opt for opt in FLAGS.beam_pipeline_options])

    builder.download_and_prepare(
            download_dir=FLAGS.download_dir,
            download_config=dl_config,
    )
    print(str(builder.info.data))

    if FLAGS.debug:
        dataset = builder.as_dataset(split=nlp.Split.TRAIN)
        pdb.set_trace()
        del dataset


def import_modules(modules):
    for module in modules:
        importlib.import_module(module)


def main(_):
    if FLAGS.module_import:
        import_modules(FLAGS.module_import)

    if FLAGS.debug_start:
        pdb.set_trace()
    if FLAGS.sleep_start:
        time.sleep(60*60*3)

    if FLAGS.disable_tqdm:
        logger.info("Disabling tqdm.")
        nlp.disable_progress_bar()

    if FLAGS.checksums_dir:
        nlp.download.add_checksums_dir(FLAGS.checksums_dir)

    datasets_to_build = set(FLAGS.datasets and FLAGS.datasets.split(","))
    datasets_to_build -= set(FLAGS.exclude_datasets.split(","))
    version = "experimental_latest" if FLAGS.experimental_latest_version else None
    logger.info("Running download_and_prepare for datasets:\n%s",
                             "\n".join(datasets_to_build))
    logger.info('Version: "%s"', version)
    builders = {
            name: nlp.builder(name, data_dir=FLAGS.data_dir, version=version)
            for name in datasets_to_build
    }

    if FLAGS.builder_config_id is not None:
        # Requesting a single config of a single dataset
        if len(builders) > 1:
            raise ValueError(
                    "--builder_config_id can only be used when building a single dataset")
        builder = builders[list(builders.keys())[0]]
        if not builder.BUILDER_CONFIGS:
            raise ValueError(
                    "--builder_config_id can only be used with datasets with configs")
        config = builder.BUILDER_CONFIGS[FLAGS.builder_config_id]
        logger.info("Running download_and_prepare for config: %s", config.name)
        builder_for_config = nlp.builder(
                builder.name, data_dir=FLAGS.data_dir, config=config, version=version)
        download_and_prepare(builder_for_config)
    else:
        for name, builder in builders.items():
            if builder.BUILDER_CONFIGS and "/" not in name:
                # If builder has multiple configs, and no particular config was
                # requested, then compute all.
                for config in builder.BUILDER_CONFIGS:
                    builder_for_config = nlp.builder(
                            builder.name, data_dir=FLAGS.data_dir, config=config,
                            version=version)
                    download_and_prepare(builder_for_config)
            else:
                # If there is a slash in the name, then user requested a specific
                # dataset configuration.
                download_and_prepare(builder)


if __name__ == "__main__":
    tf.enable_v2_behavior()
    app.run(main)



# coding=utf-8
# Copyright 2020 The HuggingFace NLP Authors and the TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""FeatureDict: Main feature connector container.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import pyarrow as pa

from .. import utils
from .feature import Tensor, FeatureConnector
from .top_level_feature import TopLevelFeature


class FeaturesDict(TopLevelFeature):
    """Composite `FeatureConnector`; each feature in `dict` has its own connector.

    The encode/decode method of the spec feature will recursively encode/decode
    every sub-connector given on the constructor.
    Other features can inherit from this class and call super() in order to get
    nested container.

    Example:

    For DatasetInfo:

    ```
    features = nlp.features.FeaturesDict({
            'input': nlp.features.Image(),
            'output': pa.int32,
    })
    ```

    At generation time:

    ```
    for image, label in generate_examples:
        yield {
                'input': image,
                'output': label
        }
    ```

    At pa.data.Dataset() time:

    ```
    for example in nlp.load(...):
        tf_input = example['input']
        tf_output = example['output']
    ```

    For nested features, the FeaturesDict will internally flatten the keys for the
    features and the conversion to pa.train.Example. Indeed, the pa.train.Example
    proto do not support nested feature, while pa.data.Dataset does.
    But internal transformation should be invisible to the user.

    Example:

    ```
    nlp.features.FeaturesDict({
            'input': pa.int32,
            'target': {
                    'height': pa.int32,
                    'width': pa.int32,
            },
    })
    ```

    Will internally store the data as:

    ```
    {
            'input': pa.io.FixedLenFeature(shape=(), dtype=pa.int32),
            'target/height': pa.io.FixedLenFeature(shape=(), dtype=pa.int32),
            'target/width': pa.io.FixedLenFeature(shape=(), dtype=pa.int32),
    }
    ```

    """

    def __init__(self, feature_dict):
        """Initialize the features.

        Args:
            feature_dict (dict): Dictionary containing the feature connectors of a
                example. The keys should correspond to the data dict as returned by
                pa.data.Dataset(). Types (pa.int32,...) and dicts will automatically
                be converted into FeatureConnector.

        Raises:
            ValueError: If one of the given features is not recognized
        """
        super(FeaturesDict, self).__init__()
        self._feature_dict = {k: to_feature(v) for k, v in feature_dict.items()}

    # Dict functions.
    # In Python 3, should inherit from collections.abc.Mapping().

    def keys(self):
        return self._feature_dict.keys()

    def items(self):
        return self._feature_dict.items()

    def values(self):
        return self._feature_dict.values()

    def __contains__(self, k):
        return k in self._feature_dict

    def __getitem__(self, key):
        """Return the feature associated with the key."""
        return self._feature_dict[key]

    def __len__(self):
        return len(self._feature_dict)

    def __iter__(self):
        return iter(self._feature_dict)

    # Feature functions

    def get_type(self) -> pa.DataType:
        """See base class for details."""
        return  pa.struct({
                feature_key: feature.get_type()
                for feature_key, feature in self._feature_dict.items()
        })

    def encode_example(self, example_dict):
        """See base class for details."""
        return {
                k: feature.encode_example(example_value)
                for k, (feature, example_value)
                in utils.zip_dict(self._feature_dict, example_dict)
        }

    def _flatten(self, x):
        """See base class for details."""
        if x and not isinstance(x, (dict, FeaturesDict)):
            raise ValueError(
                    'Error while flattening dict: FeaturesDict received a non dict item: '
                    '{}'.format(x))

        cache = {'counter': 0}  # Could use nonlocal in Python
        def _get(k):
            if x and k in x:
                cache['counter'] += 1
                return x[k]
            return None

        out = []
        for k, f in sorted(self.items()):
            out.extend(f._flatten(_get(k)))  # pylint: disable=protected-access

        if x and cache['counter'] != len(x):
            raise ValueError(
                    'Error while flattening dict: Not all dict items have been consumed, '
                    'this means that the provided dict structure does not match the '
                    '`FeatureDict`. Please check for typos in the key names. '
                    'Available keys: {}. Unrecognized keys: {}'.format(
                            list(self.keys()), list(set(x.keys()) - set(self.keys())))
            )
        return out

    def _nest(self, list_x):
        """See base class for details."""
        curr_pos = 0
        out = {}
        for k, f in sorted(self.items()):
            offset = len(f._flatten(None))  # pylint: disable=protected-access
            out[k] = f._nest(list_x[curr_pos:curr_pos+offset])  # pylint: disable=protected-access
            curr_pos += offset
        if curr_pos != len(list_x):
            raise ValueError(
                    'Error while nesting: Expected length {} does not match input '
                    'length {} of {}'.format(curr_pos, len(list_x), list_x))
        return out

    def save_metadata(self, data_dir, feature_name=None):
        """See base class for details."""
        # Recursively save all child features
        for feature_key, feature in self._feature_dict.items():
            feature_key = feature_key.replace('/', '.')
            if feature_name:
                feature_key = '-'.join((feature_name, feature_key))
            feature.save_metadata(data_dir, feature_name=feature_key)

    def load_metadata(self, data_dir, feature_name=None):
        """See base class for details."""
        # Recursively load all child features
        for feature_key, feature in self._feature_dict.items():
            feature_key = feature_key.replace('/', '.')
            if feature_name:
                feature_key = '-'.join((feature_name, feature_key))
            feature.load_metadata(data_dir, feature_name=feature_key)


class Sequence(TopLevelFeature):
    """Composite `FeatureConnector` for a `dict` where each value is a list.

    `Sequence` correspond to sequence of `nlp.features.FeatureConnector`. At
    generation time, a list for each of the sequence element is given. The output
    of `pa.data.Dataset` will batch all the elements of the sequence together.

    If the length of the sequence is static and known in advance, it should be
    specified in the constructor using the `length` param.

    Note that `Sequence` does not support features which are of type
    `pa.io.FixedLenSequenceFeature`.

    Example:
    At construction time:

    ```
    nlp.features.Sequence(nlp.features.Image(), length=NB_FRAME)
    ```

    or:

    ```
    nlp.features.Sequence({
            'frame': nlp.features.Image(shape=(64, 64, 3))
            'action': nlp.features.ClassLabel(['up', 'down', 'left', 'right'])
    }, length=NB_FRAME)
    ```

    During data generation:

    ```
    yield {
            'frame': np.ones(shape=(NB_FRAME, 64, 64, 3)),
            'action': ['left', 'left', 'up', ...],
    }
    ```

    Tensor returned by `.as_dataset()`:

    ```
    {
            'frame': pa.Tensor(shape=(NB_FRAME, 64, 64, 3), dtype=pa.uint8),
            'action': pa.Tensor(shape=(NB_FRAME,), dtype=pa.int64),
    }
    ```

    At generation time, you can specify a list of features dict, a dict of list
    values or a stacked numpy array. The lists will automatically be distributed
    into their corresponding `FeatureConnector`.

    """

    def __init__(self, feature, length=None, **kwargs):
        """Construct a sequence.

        Args:
            feature: `dict`, the features to wrap
            length: `int`, length of the sequence if static and known in advance
            **kwargs: `dict`, constructor kwargs of `nlp.features.FeaturesDict`
        """
        # Convert {} => FeaturesDict, pa.int32 => pa.int32()
        self._feature = to_feature(feature)
        self._length = length
        super(Sequence, self).__init__(**kwargs)

    @property
    def feature(self):
        """The inner feature."""
        return self._feature

    def get_type(self):
        """See base class for details."""
        # Add the additional length dimension to every serialized features
        inner_type = self._feature.get_type()
        if isinstance(inner_type, pa.StructType):
            return pa.struct(dict((f.name, pa.list_(f.type)) for f in inner_type))
        return pa.list_(inner_type)

    def encode_example(self, example_dict):
        # Convert nested dict[list] into list[nested dict]
        sequence_elements = _transpose_dict_list(example_dict)

        # If length is static, ensure that the given length match
        if self._length is not None and len(sequence_elements) != self._length:
            raise ValueError(
                    'Input sequence length do not match the defined one. Got {} != '
                    '{}'.format(len(sequence_elements), self._length)
            )

        # Empty sequences return empty arrays
        if not sequence_elements:
            def _build_empty_np(serialized_info):
                return []
                # return np.empty(
                #         shape=tuple(s if s else 0 for s in serialized_info.shape),
                #         dtype=serialized_info.dtype.as_numpy_dtype,
                # )

            return utils.map_nested(_build_empty_np, self.get_serialized_info())

        # Encode each individual elements
        sequence_elements = [
                self.feature.encode_example(sequence_elem)
                for sequence_elem in sequence_elements
        ]

        # Then convert back list[nested dict] => nested dict[list]
        def _stack_nested(sequence_elements):
            """Recursivelly stack the tensors from the same dict field."""
            if isinstance(sequence_elements[0], dict):
                return {
                        # Stack along the first dimension
                        k: _stack_nested(sub_sequence)
                        for k, sub_sequence in utils.zip_dict(*sequence_elements)
                }
            # Note: As each field can be a nested ragged list, we don't check here
            # that all elements from the list have matching dtype/shape.
            # Checking is done in `example_serializer` when elements
            # are converted to numpy array and stacked togethers.
            return list(sequence_elements)

        return _stack_nested(sequence_elements)

    def _flatten(self, x):
        """See base class for details."""
        if isinstance(x, Sequence):
            return self.feature._flatten(x.feature)  # pylint: disable=protected-access
        return self.feature._flatten(x)  # pylint: disable=protected-access

    def _nest(self, list_x):
        """See base class for details."""
        return self.feature._nest(list_x)  # pylint: disable=protected-access

    def save_metadata(self, *args, **kwargs):
        """See base class for details."""
        self._feature.save_metadata(*args, **kwargs)

    def load_metadata(self, *args, **kwargs):
        """See base class for details."""
        self._feature.load_metadata(*args, **kwargs)

    def __getitem__(self, key):
        """Convenience method to access the underlying features."""
        return self._feature[key]

    def __getattr__(self, key):
        """Allow to access the underlying attributes directly."""
        return getattr(self._feature, key)

    # The __getattr__ method triggers an infinite recursion loop when loading a
    # pickled instance. So we override that name in the instance dict, and remove
    # it when unplickling.
    def __getstate__(self):
        state = self.__dict__.copy()
        state['__getattr__'] = 0
        return state

    def __setstate__(self, state):
        del state['__getattr__']
        self.__dict__.update(state)


def _np_to_list(elem):
    """Returns list from list, tuple or ndarray."""
    if isinstance(elem, list):
        return elem
    elif isinstance(elem, tuple):
        return list(elem)
    elif isinstance(elem, np.ndarray):
        return list(elem)
    else:
        raise ValueError(
                'Input elements of a sequence should be either a numpy array, a '
                'python list or tuple. Got {}'.format(type(elem)))


def _transpose_dict_list(dict_list):
    """Transpose a nested dict[list] into a list[nested dict]."""
    # 1. Unstack numpy arrays into list
    dict_list = utils.map_nested(_np_to_list, dict_list, dict_only=True)

    # 2. Extract the sequence length (and ensure the length is constant for all
    # elements)
    length = {'value': None}  # dict because `nonlocal` is Python3 only
    def update_length(elem):
        if length['value'] is None:
            length['value'] = len(elem)
        elif length['value'] != len(elem):
            raise ValueError(
                    'The length of all elements of one sequence should be the same. '
                    'Got {} != {}'.format(length['value'], len(elem)))
        return elem
    utils.map_nested(update_length, dict_list, dict_only=True)

    # 3. Extract each individual elements
    return [
            utils.map_nested(lambda elem: elem[i], dict_list, dict_only=True)   # pylint: disable=cell-var-from-loop
            for i in range(length['value'])
    ]


def to_feature(value):
    """Convert the given value to Feature if necessary."""
    if isinstance(value, FeatureConnector):
        return value
    elif isinstance(value, dict):
        return FeaturesDict(value)
    return Tensor(shape=(), dtype=value)



# coding=utf-8
# Copyright 2020 The HuggingFace NLP Authors and the TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""`nlp.features.FeatureConnector` API defining feature types."""

from .class_label_feature import ClassLabel
from .feature import FeatureConnector
from .feature import Tensor
from .feature import TensorInfo
from .features_dict import FeaturesDict, Sequence
from .text_feature import Text
from .translation_feature import Translation
from .translation_feature import TranslationVariableLanguages

__all__ = [
    "ClassLabel",
    "FeatureConnector",
    "FeaturesDict",
    "Tensor",
    "TensorInfo",
    "Sequence",
    "Text",
]
